#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Mocked vendor server for provision via a hacked aiohttp."""
from __future__ import annotations

from http import HTTPMethod, HTTPStatus
import re
from typing import TYPE_CHECKING

from evohomeasync2.const import AUTH_URL, URL_BASE
from evohomeasync2.schema import vol  # voluptuous
from evohomeasync2.schema import convert_to_get_schedule
from evohomeasync2.schema.const import (
    SZ_DHW,
    SZ_DHW_ID,
    SZ_DOMESTIC_HOT_WATER,
    SZ_GATEWAYS,
    SZ_LOCATION,
    SZ_LOCATION_ID,
    SZ_SYSTEM_ID,
    SZ_TEMPERATURE_CONTROL_SYSTEM,
    SZ_TEMPERATURE_CONTROL_SYSTEMS,
    SZ_TEMPERATURE_ZONE,
    SZ_ZONE_ID,
    SZ_ZONES,
)
from evohomeasync2.schema.schedule import (
    SCH_PUT_SCHEDULE_DHW,
    SCH_PUT_SCHEDULE_ZONE,
)

from .const import (
    MOCK_AUTH_RESPONSE,
    MOCK_FULL_CONFIG,
    MOCK_LOCN_STATUS,
    MOCK_SCHEDULE_DHW,
    MOCK_SCHEDULE_ZONE,
)
from .const import user_config_from_full_config as _user_config_from_full_config


if TYPE_CHECKING:
    from evohomeasync2.schema import (
        _DhwIdT,
        _LocationIdT,
        _SystemIdT,
        _UserIdT,
        _ZoneIdT,
    )
    from .const import _bodyT, _methodT, _statusT, _urlT


def _dhw_id(url: _urlT) -> _DhwIdT:
    """Extract a DHW ID from a URL."""
    return url.split(f"{SZ_DOMESTIC_HOT_WATER}/")[1].split("/")[0]


def _loc_id(url: _urlT) -> _LocationIdT:
    """Extract a Location ID from a URL."""
    return url.split(f"{SZ_LOCATION}/")[1].split("/")[0]


def _tcs_id(url: _urlT) -> _SystemIdT:
    """Extract a TCS ID from a URL."""
    return url.split(f"{SZ_TEMPERATURE_CONTROL_SYSTEM}/")[1].split("/")[0]


def _usr_id(url: _urlT) -> _UserIdT:
    """Extract a TCS ID from a URL."""
    return url.split("?userId=")[1].split("&")[0]


def _zon_id(url: _urlT) -> _ZoneIdT:
    """Extract a Zone ID from a URL."""
    return url.split(f"{SZ_TEMPERATURE_ZONE}/")[1].split("/")[0]


def validate_id_of_url(id_fnc):
    """Validate the ID in the URL and set the status accordingly."""

    def decorator(func):
        def wrapper(self: MockedServer) -> _bodyT:
            if self._method != HTTPMethod.GET:
                self.status = HTTPStatus.METHOD_NOT_ALLOWED
                return {"message": "Method not allowed"}

            try:
                id: str = id_fnc(self._url)
            except IndexError:
                self.status = HTTPStatus.NOT_FOUND
                return {"message": "Not Found"}

            if not id.isdigit():
                self.status = HTTPStatus.BAD_REQUEST
                return [{"message": "Bad request"}]

            if result := func(self):
                return result

            self.status = HTTPStatus.UNAUTHORIZED
            return [{"message": "Unauthorized"}]

        return wrapper

    return decorator


class MockedServer:
    """Mocked vendor server for provision via a hacked aiohttp."""

    def __init__(
        self,
        full_config: dict,
        locn_status: dict,
        /,
        *,
        zone_schedule: dict | None = None,
        dhw_schedule: dict | None = None,
    ) -> None:
        self._full_config = full_config or MOCK_FULL_CONFIG
        self._locn_status = locn_status or MOCK_LOCN_STATUS
        self._zon_schedule = zone_schedule or MOCK_SCHEDULE_ZONE
        self._dhw_schedule = dhw_schedule or MOCK_SCHEDULE_DHW

        self._schedules = {}
        self._user_config = self._user_config_from_full_config(self._full_config)

        self.body: _bodyT | None = None
        self._method: _methodT | None = None
        self.status: _statusT | None = None
        self._url: _urlT | None = None

    def request(
        self, method: _methodT, url: _urlT, data: dict | str | None = None
    ) -> _bodyT:
        self._method = method
        self._url = url
        self._data = data

        self.body, self.status = None, None
        for pattern, method in REQUEST_MAP.items():
            if re.search(pattern, url):
                self.body: _bodyT | None = method(self)
                break
        else:
            self.status = HTTPStatus.NOT_FOUND
            return """
                <html>
                    <head><title>Error</title></head>
                    <body><h1>Not Found</h1></body>
                </html>
            """

        if not self.status:
            self.status = HTTPStatus.OK if self.body else HTTPStatus.NOT_FOUND
        return self.body

    def oauth_token(self) -> _bodyT | None:
        if self._method != HTTPMethod.POST:
            self.status = HTTPStatus.METHOD_NOT_ALLOWED
        elif self._url == AUTH_URL:
            return MOCK_AUTH_RESPONSE

    def usr_account(self) -> _bodyT | None:
        if self._method != HTTPMethod.GET:
            self.status = HTTPStatus.METHOD_NOT_ALLOWED
            return {"message": "Method not allowed"}

        elif self._url == f"{URL_BASE}/userAccount":
            return self._user_config

        else:
            self.status = HTTPStatus.NOT_FOUND
            return {"message": "Not found"}

    @validate_id_of_url(_usr_id)
    def all_config(self) -> _bodyT | None:  # full_locn
        usr_id = _usr_id(self._url)

        if self._user_config["userId"] == usr_id:
            return self._full_config

    def loc_config(self) -> _bodyT | None:
        raise NotImplementedError

    @validate_id_of_url(_loc_id)
    def loc_status(self) -> _bodyT:
        loc_id = _loc_id(self._url)

        if self._locn_status[SZ_LOCATION_ID] == loc_id:
            return self._locn_status

    def tcs_mode(self) -> _bodyT | None:
        raise NotImplementedError

    @validate_id_of_url(_tcs_id)
    def tcs_status(self) -> _bodyT:
        tcs_id = _tcs_id(self._url)

        for gwy in self._locn_status[SZ_GATEWAYS]:
            for tcs in gwy[SZ_TEMPERATURE_CONTROL_SYSTEMS]:
                if tcs[SZ_SYSTEM_ID] == tcs_id:
                    return tcs

    def zon_schedule(self) -> _bodyT | None:
        zon_id = _zon_id(self._url)

        if self._method == HTTPMethod.GET:
            return self._schedules.get(zon_id, self._zon_schedule)

        if self._method != HTTPMethod.PUT:
            self.status = HTTPStatus.METHOD_NOT_ALLOWED
            return {"message": "Method not allowed"}

        if not isinstance(self._data, dict):
            self.status = HTTPStatus.BAD_REQUEST
            return [{"message": "Bad Request (invalid schedule: not a dict)"}]

        try:
            SCH_PUT_SCHEDULE_ZONE(self._data)
        except vol.Invalid:
            self.status = HTTPStatus.BAD_REQUEST
            return {"message": "Bad Request (invalid schedule: invalid schema)"}

        self._schedules[zon_id] = convert_to_get_schedule(self._data)
        return {"id": "1234567890"}

    def zon_mode(self) -> _bodyT | None:
        raise NotImplementedError

    @validate_id_of_url(_zon_id)
    def zon_status(self) -> _bodyT:
        zon_id = _zon_id(self._url)

        for gwy in self._locn_status[SZ_GATEWAYS]:
            for tcs in gwy[SZ_TEMPERATURE_CONTROL_SYSTEMS]:
                for zone in tcs[SZ_ZONES]:
                    if zone[SZ_ZONE_ID] == zon_id:
                        return zone

    def dhw_schedule(self) -> _bodyT | None:
        dhw_id = _dhw_id(self._url)

        if self._method == HTTPMethod.GET:
            return self._schedules.get(dhw_id, self._dhw_schedule)

        if self._method != HTTPMethod.PUT:
            self.status = HTTPStatus.METHOD_NOT_ALLOWED
            return {"message": "Method not allowes"}

        if not isinstance(self._data, dict):
            self.status = HTTPStatus.BAD_REQUEST
            return [{"message": "Bad Request (invalid schedule: not a dict)"}]

        try:
            SCH_PUT_SCHEDULE_DHW(self._data)
        except vol.Invalid:
            self.status = HTTPStatus.BAD_REQUEST
            return {"message": "Bad Request (invalid schedule: invalid schema)"}

        self._schedules[dhw_id] = convert_to_get_schedule(self._data)
        return {"id": "1234567890"}

    @validate_id_of_url(_dhw_id)
    def dhw_status(self) -> _bodyT:
        dhw_id = _dhw_id(self._url)

        for gwy in self._locn_status[SZ_GATEWAYS]:
            for tcs in gwy[SZ_TEMPERATURE_CONTROL_SYSTEMS]:
                if dhw := tcs.get(SZ_DHW):
                    if dhw[SZ_DHW_ID] == dhw_id:
                        return dhw

    def dhw_mode(self) -> _bodyT | None:
        raise NotImplementedError

    #

    def OUT__response_for_request(self) -> _bodyT | None:
        """Set the response body (and status) according to the request."""

        if "/Auth" in self._url and self._method == HTTPMethod.POST:
            # POST /Auth/OAuth/Token
            return self.auth_response()

        if "/userAccount" in self._url and self._method == HTTPMethod.GET:
            # GET /userAccount
            return self.user_config()

        if "/location" in self._url and self._method == HTTPMethod.GET:
            return self._response_for_location_request()

        if "/temperatureControlSystem" in self._url and self._method == HTTPMethod.PUT:
            if self._method != HTTPMethod.PUT:
                pass

            elif re.search(r"temperatureControlSystem/.*/mode", self._url):
                # PUT /temperatureControlSystem/{systemId}/mode"
                return self.full_config()

        if "/temperatureZone" in self._url and self._method in (
            HTTPMethod.GET,
            HTTPMethod.PUT,
        ):
            return self._response_for_zone_request()

        if "/domesticHotWater" in self._url and self._method in (
            HTTPMethod.GET,
            HTTPMethod.PUT,
        ):
            return self._response_for_dhw_request()

    def OUT_response_for_location_request(self) -> _bodyT | None:
        """"""

        if self._method != HTTPMethod.GET:
            pass

        elif re.search(r"location/installationInfo", self._url):
            # GET /location/installationInfo?userId={userId}
            return self.full_config()

        elif re.search(r"location/.*/installationInfo", self._url):
            # GET /location/{locationId}/installationInfo
            return self.locn_config(location_id=None)

        elif re.search(r"location/.*/status", self._url):
            # GET /location/{locationId}/status
            return self.locn_status(location_id=None)

    def OUT_response_for_zone_request(self) -> _bodyT | None:
        """"""

        def zone_id() -> _ZoneIdT:
            return self._url.split("temperatureZone/")[1].split("/")[0]

        if self._method == HTTPMethod.GET:
            if re.search(r"temperatureZone/.*/schedule", self._url):
                # GET /temperatureZone/{zoneId}/schedule  # /{zone_type}/{zoneId}/schedule
                return self.zone_schedule(zone_id=zone_id())

            if re.search(r"temperatureZone/.*/status", self._url):
                # GET /temperatureZone/{zoneId}/status  # /{zone_type}/{zoneId}/schedule
                return self.zone_status(zone_id=zone_id())

        elif self._method == HTTPMethod.PUT:
            if re.search(r"temperatureZone/.*/schedule", self._url):
                # PUT /temperatureZone/{zoneId}/schedule
                return {"id": "123456789"}

            if re.search(r"temperatureZone/.*/heatSetpoint", self._url):  # aka mode
                # PUT /temperatureZone/{zoneId}/heatSetpoint
                raise NotImplementedError

    def OUT_response_for_dhw_request(self) -> _bodyT | None:
        """"""

        def dhw_id() -> _DhwIdT:
            return self._url.split("domesticHotWater/")[1].split("/")[0]

        if self._method == HTTPMethod.GET:
            if re.search(r"domesticHotWater/.*/schedule", self._url):
                # GET /domesticHotWater/{dhwId}/schedule  # /{zone_type}/{zoneId}/schedule
                return self.dhw_schedule(dhw_id=dhw_id())

            if re.search(r"domesticHotWater/.*/status", self._url):
                # GET /domesticHotWater/{dhwId}/status
                return self.dhw_status(dhw_id=dhw_id())

        elif self._method == HTTPMethod.PUT:
            if re.search(r"domesticHotWater/.*/schedule", self._url):
                # PUT /domesticHotWater/{dhwId}/schedule
                return {"id": "123456789"}

            if re.search(r"domesticHotWater/.*/state", self._url):  # aka mode
                # PUT /domesticHotWater/{dhwId}/state
                raise NotImplementedError

    def OLD_dhw_status_old(self, dhw_id: _DhwIdT) -> dict | None:
        for gwy in self._locn_status[SZ_GATEWAYS]:
            for tcs in gwy[SZ_TEMPERATURE_CONTROL_SYSTEMS]:
                if (dhw := tcs.get(SZ_DHW)) and dhw[SZ_DHW_ID] == dhw_id:
                    return dhw

        self.status = 404

    def OUT_zone_status(self, zone_id: _ZoneIdT) -> dict | None:
        for gwy in self._locn_status[SZ_GATEWAYS]:
            for tcs in gwy[SZ_TEMPERATURE_CONTROL_SYSTEMS]:
                for zone in tcs[SZ_ZONES]:
                    if zone[SZ_ZONE_ID] == zone_id:
                        return zone

        self.status = 404

    #

    @staticmethod
    def _user_config_from_full_config(full_config: list) -> dict:
        """Create a valid MOCK_USER_CONFIG from a MOCK_FULL_CONFIG."""
        return _user_config_from_full_config(full_config)


REQUEST_MAP = {
    #
    r"/Auth/OAuth/Token": MockedServer.oauth_token,
    #
    r"/userAccount$": MockedServer.usr_account,
    r"/location/installationInfo": MockedServer.all_config,
    #
    r"/location/.*/installationInfo": MockedServer.loc_config,
    r"/location/.*/status": MockedServer.loc_status,
    #
    r"/temperatureControlSystem/.*/mode": MockedServer.tcs_mode,
    r"/temperatureControlSystem/.*/status": MockedServer.tcs_status,
    #
    r"/temperatureZone/.*/status": MockedServer.zon_status,
    r"/temperatureZone/.*/heatSetpoint": MockedServer.zon_mode,
    r"/temperatureZone/.*/schedule": MockedServer.zon_schedule,
    #
    r"/domesticHotWater/.*/status": MockedServer.dhw_status,
    r"/domesticHotWater/.*/state": MockedServer.dhw_mode,
    r"/domesticHotWater/.*/schedule": MockedServer.dhw_schedule,
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Provides handling of TCC temperature control systems."""

# TODO: add provision for cooling
# TODO: add set_mode() for non-evohome modes (e.g. "Heat", "Off")

from __future__ import annotations

from datetime import datetime as dt
import json
from typing import TYPE_CHECKING, Final, NoReturn

from .const import API_STRFTIME, SystemMode
from .exceptions import DeprecationError, InvalidParameter, InvalidSchema
from .hotwater import HotWater
from .schema import SCH_TCS_STATUS
from .schema.const import (
    SZ_ACTIVE_FAULTS,
    SZ_ALLOWED_SYSTEM_MODES,
    SZ_DHW,
    SZ_IS_AVAILABLE,
    SZ_MODEL_TYPE,
    SZ_PERMANENT,
    SZ_SYSTEM_ID,
    SZ_SYSTEM_MODE,
    SZ_SYSTEM_MODE_STATUS,
    SZ_TARGET_HEAT_TEMPERATURE,
    SZ_TEMPERATURE,
    SZ_TEMPERATURE_CONTROL_SYSTEM,
    SZ_TIME_UNTIL,
    SZ_ZONES,
)
from .zone import Zone


if TYPE_CHECKING:
    import logging

    from . import Broker, Gateway, Location
    from .schema import _DhwIdT, _EvoDictT, _EvoListT, _FilePathT, _SystemIdT, _ZoneIdT


# used by temperatures() and *_schedules()...
SZ_ID = "id"
SZ_NAME = "name"
SZ_TEMP = "temp"
SZ_THERMOSTAT = "thermostat"
SZ_SCHEDULE = "schedule"
SZ_SETPOINT = "setpoint"


class _ControlSystemDeprecated:
    """Deprecated attributes and methods removed from the evohome-client namespace."""

    async def set_status_reset(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "ControlSystem.set_status_reset() is deprecrated, use .reset_mode()"
        )

    async def set_status(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "ControlSystem.set_status() is deprecrated, use .set_mode()"
        )

    async def set_status_normal(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "ControlSystem.set_status_normal() is deprecrated, use .set_auto()"
        )

    async def set_status_away(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "ControlSystem.set_status_away() is deprecrated, use .set_away()"
        )

    async def set_status_custom(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "ControlSystem.set_status_custom() is deprecrated, use .set_custom()"
        )

    async def set_status_dayoff(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "ControlSystem.set_status_dayoff() is deprecrated, use .set_dayoff()"
        )

    async def set_status_eco(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "ControlSystem.set_status_eco() is deprecrated, use .set_eco()"
        )

    async def set_status_heatingoff(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "ControlSystem.set_status_heatingoff() is deprecrated, use .set_heatingoff()"
        )

    async def zone_schedules_backup(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "TCS.zone_schedules_backup() is deprecated, use .backup_schedules()"
        )

    async def zone_schedules_restore(self, *args, **kwargs) -> NoReturn:
        raise DeprecationError(
            "TCS.zone_schedules_restore() is deprecated, use .restore_schedules()"
        )


class ControlSystem(_ControlSystemDeprecated):
    """Instance of a gateway's TCS (temperatureControlSystem)."""

    STATUS_SCHEMA: Final = SCH_TCS_STATUS
    TYPE: Final[str] = SZ_TEMPERATURE_CONTROL_SYSTEM

    def __init__(self, gateway: Gateway, config: _EvoDictT) -> None:
        self.gateway = gateway
        self.location: Location = gateway.location

        self._broker: Broker = gateway._broker
        self._logger: logging.Logger = gateway._logger

        self._status: _EvoDictT = {}
        self._config: Final[_EvoDictT] = {
            k: v for k, v in config.items() if k not in (SZ_DHW, SZ_ZONES)
        }

        try:
            assert self.systemId, "Invalid config dict"
        except AssertionError as exc:
            raise InvalidSchema(str(exc))
        self._id = self.systemId

        self._zones: list[Zone] = []
        self.zones: dict[str, Zone] = {}  # zone by name! what to do if name changed?
        self.zones_by_id: dict[str, Zone] = {}
        self.hotwater: None | HotWater = None

        dhw_config: _EvoDictT
        zon_config: _EvoDictT

        for zon_config in config[SZ_ZONES]:
            zone = Zone(self, zon_config)

            self._zones.append(zone)
            self.zones[zone.name] = zone
            self.zones_by_id[zone.zoneId] = zone

        if dhw_config := config.get(SZ_DHW):  # type: ignore[assignment]
            self.hotwater = HotWater(self, dhw_config)

    def __str__(self) -> str:
        return f"{self._id} ({self.TYPE})"

    @property
    def allowedSystemModes(self) -> _EvoListT:
        return self._config[SZ_ALLOWED_SYSTEM_MODES]

    @property
    def modelType(self) -> str:
        return self._config[SZ_MODEL_TYPE]

    @property
    def systemId(self) -> _SystemIdT:
        return self._config[SZ_SYSTEM_ID]

    async def _refresh_status(self) -> None:
        await self.location.refresh_status()

    def _update_status(self, tcs_status: _EvoDictT) -> None:
        self._status = tcs_status

    @property
    def activeFaults(self) -> _EvoListT | None:
        return self._status.get(SZ_ACTIVE_FAULTS)

    @property
    def systemModeStatus(self) -> _EvoDictT | None:
        return self._status.get(SZ_SYSTEM_MODE_STATUS)

    async def _set_mode(self, mode: dict) -> None:
        """Set the TCS mode."""  # {'mode': 'Auto', 'isPermanent': True}
        _ = await self._broker.put(f"{self.TYPE}/{self._id}/mode", json=mode)

    async def reset_mode(self) -> None:
        """Set the TCS to auto mode (and DHW/all zones to FollowSchedule mode)."""
        await self.set_status(SystemMode.AUTO_WITH_RESET)

    async def set_mode(self, mode: SystemMode, /, *, until: dt | None = None) -> None:
        """Set the system to a mode, either indefinitely, or for a set time."""

        request: _EvoDictT

        if mode not in [m[SZ_SYSTEM_MODE] for m in self.allowedSystemModes]:
            raise InvalidParameter(f"Unsupported/unknown mode: {mode}")

        if until is None:
            request = {
                SZ_SYSTEM_MODE: mode,
                SZ_PERMANENT: True,
                SZ_TIME_UNTIL: None,
            }
        else:
            request = {
                SZ_SYSTEM_MODE: mode,
                SZ_PERMANENT: False,
                SZ_TIME_UNTIL: until.strftime(API_STRFTIME),
            }

        await self._set_mode(request)

    async def set_auto(self) -> None:
        """Set the system into normal mode."""
        await self.set_status(SystemMode.AUTO)

    async def set_away(self, /, *, until: dt | None = None) -> None:
        """Set the system into away mode."""
        await self.set_status(SystemMode.AWAY, until=until)

    async def set_custom(self, /, *, until: dt | None = None) -> None:
        """Set the system into custom mode."""
        await self.set_status(SystemMode.CUSTOM, until=until)

    async def set_dayoff(self, /, *, until: dt | None = None) -> None:
        """Set the system into dayoff mode."""
        await self.set_status(SystemMode.DAY_OFF, until=until)

    async def set_eco(self, /, *, until: dt | None = None) -> None:
        """Set the system into eco mode."""
        await self.set_status(SystemMode.AUTO_WITH_ECO, until=until)

    async def set_heatingoff(self, /, *, until: dt | None = None) -> None:
        """Set the system into heating off mode."""
        await self.set_status(SystemMode.HEATING_OFF, until=until)

    async def temperatures(self) -> _EvoListT:
        """A convienience function to return the latest temperatures and setpoints."""

        await self.location.refresh_status()

        result = []

        if dhw := self.hotwater:
            dhw_status = {
                SZ_THERMOSTAT: "DOMESTIC_HOT_WATER",
                SZ_ID: dhw.dhwId,
                SZ_NAME: dhw.name,
                SZ_TEMP: None,
            }

            if (
                isinstance(dhw.temperatureStatus, dict)
                and dhw.temperatureStatus[SZ_IS_AVAILABLE]
            ):
                dhw_status[SZ_TEMP] = dhw.temperatureStatus[SZ_TEMPERATURE]

            result.append(dhw_status)

        for zone in self._zones:
            zone_status = {
                SZ_THERMOSTAT: "EMEA_ZONE",
                SZ_ID: zone.zoneId,
                SZ_NAME: zone.name,
                SZ_SETPOINT: None,
                SZ_TEMP: None,
            }

            if isinstance(zone.setpointStatus, dict):
                zone_status[SZ_SETPOINT] = zone.setpointStatus[
                    SZ_TARGET_HEAT_TEMPERATURE
                ]

            if (
                isinstance(zone.temperatureStatus, dict)
                and zone.temperatureStatus[SZ_IS_AVAILABLE]
            ):
                zone_status[SZ_TEMP] = zone.temperatureStatus[SZ_TEMPERATURE]

            result.append(zone_status)

        return result

    async def backup_schedules(self, filename: _FilePathT) -> None:
        """Backup all schedules from the control system to the file."""

        self._logger.info(
            f"Backing up schedules from {self.systemId} ({self.location.name})"
            f", to {filename}"
        )

        schedules = {}

        for zone in self._zones:
            schedule = await zone.get_schedule()
            schedules[zone.zoneId] = {
                SZ_NAME: zone.name,
                SZ_SCHEDULE: schedule,
            }

        if self.hotwater:
            schedule = await self.hotwater.get_schedule()
            schedules[self.hotwater.dhwId] = {
                SZ_NAME: self.hotwater.name,
                SZ_SCHEDULE: schedule,
            }

        with open(filename, "w") as file_output:
            file_output.write(json.dumps(schedules, indent=4))

        self._logger.info("Backup completed.")

    async def restore_schedules(
        self, filename: _FilePathT, match_by_name: bool = False
    ) -> None:
        """Restore all schedules from the file to the TCS.

        The default is to match a schedule to its zone/dhw by id.
        """

        async def restore_by_id(id: _ZoneIdT | _DhwIdT, schedule: dict) -> bool:
            """Restore schedule by id and return False if there was no match."""

            name = schedule.get(SZ_NAME)

            if self.hotwater and self.hotwater.dhwId == id:
                await self.hotwater.set_schedule(json.dumps(schedule[SZ_SCHEDULE]))

            elif zone := self.zones_by_id.get(id):
                await zone.set_schedule(json.dumps(schedule[SZ_SCHEDULE]))

            else:
                self._logger.warning(
                    f"Ignoring schedule of {id} ({name}): unknown id"
                    ", consider matching by name rather than by id"
                )
                return False

            return True

        async def restore_by_name(id: _ZoneIdT | _DhwIdT, schedule: dict) -> bool:
            """Restore schedule by name and return False if there was no match."""

            name = schedule[SZ_NAME]  # don't use .get()

            if self.hotwater and name == self.hotwater.name:
                await self.hotwater.set_schedule(json.dumps(schedule[SZ_SCHEDULE]))

            elif zone := self.zones.get(name):
                await zone.set_schedule(json.dumps(schedule[SZ_SCHEDULE]))

            else:
                self._logger.warning(
                    f"Ignoring schedule of {id} ({name}): unknown name"
                    ", consider matching by id rather than by name"
                )
                return False

            return True

        self._logger.info(
            f"Restoring schedules (matched by {'name' if match_by_name else 'id'}) to "
            f"{self.systemId} ({self.location.name}), from {filename}"
        )

        with open(filename, "r") as file_input:
            schedule_db = file_input.read()
            schedules: dict = json.loads(schedule_db)

        with_errors = False

        for id, schedule in schedules.items():
            assert isinstance(schedule, dict)  # mypy

            if match_by_name:
                matched = await restore_by_name(id, schedule)
            else:
                matched = await restore_by_id(id, schedule)

            with_errors = with_errors and not matched

        if with_errors or len(schedules) != len(self.zones) + 1 if self.hotwater else 0:
            self._logger.warning("Restore completed, but with unmatched schedules.")
        else:
            self._logger.info("Restore completed.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohome-async - validate the handling of vendor APIs (URLs)."""
from __future__ import annotations


from datetime import datetime as dt
from datetime import timedelta as td
from http import HTTPMethod, HTTPStatus
import pytest
import pytest_asyncio

import evohomeasync2 as evo
from evohomeasync2 import Location, Gateway, ControlSystem
from evohomeasync2.const import API_STRFTIME, DhwState, ZoneMode
from evohomeasync2.schema.const import (
    SZ_MODE,
    SZ_STATE,
    SZ_STATE_STATUS,
    SZ_UNTIL,
    SZ_UNTIL_TIME,
)
from evohomeasync2.schema.helpers import pascal_case

from . import _DEBUG_USE_REAL_AIOHTTP
from .helpers import aiohttp, extract_oauth_tokens  # aiohttp may be mocked
from .helpers import user_credentials as _user_credentials
from .helpers import client_session as _client_session
from .helpers import should_work, should_fail, wait_for_comm_task


_global_oauth_tokens: tuple[str, str, dt] = None, None, None


@pytest.fixture()
def user_credentials():
    return _user_credentials()


@pytest_asyncio.fixture
async def session():
    client_session = _client_session()
    try:
        yield client_session
    finally:
        await client_session.close()


@pytest.fixture(autouse=True)
def patches_for_tests(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("evohomeasync2.broker.aiohttp", aiohttp)


async def instantiate_client(
    username: str,
    password: str,
    session: aiohttp.ClientSession | None = None,
):
    """Instantiate a client, and logon to the vendor API."""

    global _global_oauth_tokens

    refresh_token, access_token, access_token_expires = _global_oauth_tokens

    # Instantiation, NOTE: No API calls invoked during instantiation
    client = evo.EvohomeClient(
        username,
        password,
        session=session,
        refresh_token=refresh_token,
        access_token=access_token,
        access_token_expires=access_token_expires,
    )

    # Authentication
    await client._broker._basic_login()
    _global_oauth_tokens = extract_oauth_tokens(client)

    return client


async def _test_task_id(
    username: str,
    password: str,
    session: aiohttp.ClientSession,
) -> None:
    """Test the task_id returned when using the vendor's RESTful APIs.

    This test can be used to prove that JSON keys are can be camelCase or PascalCase.
    """

    loc: Location
    gwy: Gateway
    tcs: ControlSystem

    client = await instantiate_client(username, password, session=session)
    _ = await client.user_account()
    _ = await client._installation(refresh_status=False)

    for loc in client.locations:
        for gwy in loc._gateways:
            for tcs in gwy._control_systems:
                if tcs.hotwater:
                    # if (dhw := tcs.hotwater) and dhw.temperatureStatus['isAvailable']:
                    dhw = tcs.hotwater
                    break
    # else:
    #     pytest.skip("No available DHW found")
    #

    GET_URL = f"{dhw.TYPE}/{dhw._id}/status"
    PUT_URL = f"{dhw.TYPE}/{dhw._id}/state"

    #
    # PART 0: Get initial state...
    old_status = await should_work(client, HTTPMethod.GET, GET_URL)  # HTTP 200
    # {
    #     'dhwId': '3933910',
    #     'temperatureStatus': {'isAvailable': False},
    #     'stateStatus': {'state': 'Off', 'mode': 'FollowSchedule'},
    #     'activeFaults': []
    # }  # HTTP 200
    # {
    #     'dhwId': '3933910',
    #     'temperatureStatus': {'temperature': 21.0, 'isAvailable': True},
    #     'stateStatus': {
    #         'state': 'On',
    #         'mode': 'TemporaryOverride',
    #         'until': '2023-10-30T18:40:00Z'
    #     },
    #     'activeFaults': []
    # }  # HTTP 200

    old_mode = {
        SZ_MODE: old_status[SZ_STATE_STATUS][SZ_MODE],
        SZ_STATE: old_status[SZ_STATE_STATUS][SZ_STATE],
        SZ_UNTIL_TIME: old_status[SZ_STATE_STATUS].get(SZ_UNTIL),
    }  # NOTE: untilTime/until

    #
    # PART 1: Try the basic functionality...
    # new_mode = {SZ_MODE: ZoneMode.PERMANENT_OVERRIDE, SZ_STATE: DhwState.OFF, SZ_UNTIL_TIME: None}
    new_mode = {
        SZ_MODE: ZoneMode.TEMPORARY_OVERRIDE,
        SZ_STATE: DhwState.ON,
        SZ_UNTIL_TIME: (dt.now() + td(hours=1)).strftime(API_STRFTIME),
    }

    result = await should_work(client, HTTPMethod.PUT, PUT_URL, json=new_mode)
    # {'id': '840367013'}  # HTTP 201/Created

    task_id = result[0]["id"] if isinstance(result, list) else result["id"]
    url_tsk = f"commTasks?commTaskId={task_id}"

    _ = await should_work(client, HTTPMethod.GET, url_tsk)
    # {'commtaskId': '840367013', 'state': 'Created'}
    # {'commtaskId': '840367013', 'state': 'Succeeded'}

    # dtm = dt.now()
    _ = await wait_for_comm_task(client, task_id, timeout=3)
    # assert (dt.now() - dtm).total_seconds() < 2

    #
    # PART 2A: Try different capitalisations of the JSON keys...
    new_mode = {
        SZ_MODE: ZoneMode.TEMPORARY_OVERRIDE,
        SZ_STATE: DhwState.ON,
        SZ_UNTIL_TIME: (dt.now() + td(hours=1)).strftime(API_STRFTIME),
    }
    _ = await should_work(client, HTTPMethod.PUT, PUT_URL, json=new_mode)  # HTTP 201
    _ = await wait_for_comm_task(client, task_id, timeout=3)
    status = await should_work(client, HTTPMethod.GET, GET_URL)

    new_mode = {  # NOTE: different capitalisation, until time
        pascal_case(SZ_MODE): ZoneMode.TEMPORARY_OVERRIDE,
        pascal_case(SZ_STATE): DhwState.ON,
        pascal_case(SZ_UNTIL_TIME): (dt.now() + td(hours=2)).strftime(API_STRFTIME),
    }
    _ = await should_work(client, HTTPMethod.PUT, PUT_URL, json=new_mode)
    _ = await wait_for_comm_task(client, task_id)
    status = await should_work(client, HTTPMethod.GET, GET_URL)

    #
    # PART 3: Restore the original mode
    _ = await should_work(client, HTTPMethod.PUT, PUT_URL, json=old_mode)
    _ = await wait_for_comm_task(client, task_id)
    status = await should_work(client, HTTPMethod.GET, GET_URL)

    assert status  # == old_status

    #
    # PART 4A: Try bad JSON...
    bad_mode = {
        SZ_STATE: ZoneMode.TEMPORARY_OVERRIDE,
        SZ_MODE: DhwState.OFF,
        SZ_UNTIL_TIME: None,
    }
    _ = await should_fail(
        client, HTTPMethod.PUT, PUT_URL, json=bad_mode, status=HTTPStatus.BAD_REQUEST
    )  #
    # x = [{
    #     "code": "InvalidInput", "message": """
    #         Error converting value 'TemporaryOverride'
    #         to type 'DomesticHotWater.Enums.EMEADomesticHotWaterState'.
    #         Path 'state', line 1, position 29.
    #     """
    # }, {
    #     "code": "InvalidInput", "message": """
    #         Error converting value 'Off'
    #         to type 'DomesticHotWater.Enums.EMEADomesticHotWaterSetpointMode'.
    #         Path 'mode', line 1, position 44.
    #     """
    # }]  # NOTE: message has been slightly edited for readability

    #
    # PART 4B: Try 'bad' task_id values...
    url_tsk = "commTasks?commTaskId=ABC"
    _ = await should_fail(
        client, HTTPMethod.GET, url_tsk, status=HTTPStatus.BAD_REQUEST
    )  # [{"code": "InvalidInput", "message": "Invalid Input."}]

    url_tsk = "commTasks?commTaskId=12345678"
    _ = await should_fail(
        client, HTTPMethod.GET, url_tsk, status=HTTPStatus.NOT_FOUND
    )  # [{"code": "CommTaskNotFound", "message": "Communication task not found."}]

    pass


@pytest.mark.asyncio
async def test_task_id(
    user_credentials: tuple[str, str], session: aiohttp.ClientSession
) -> None:
    """Test /location/{locationId}/status"""

    if not _DEBUG_USE_REAL_AIOHTTP:
        pytest.skip("Test is only valid with a real server")

    try:
        await _test_task_id(*user_credentials, session)
    except evo.AuthenticationFailed:
        pytest.skip("Unable to authenticate")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync provides an async client for the *original* Evohome API.

It is a faithful async port of https://github.com/watchforstock/evohome-client

Further information at: https://evohome-client.readthedocs.io
"""
from __future__ import annotations

from .client import EvohomeClient  # noqa: F401
from .exceptions import (  # noqa: F401
    AuthenticationFailed,
    EvohomeError,
    InvalidSchema,
    RateLimitExceeded,
    RequestFailed,
)


__version__ = "0.4.5"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohome-async - validate the config & status schemas."""
from __future__ import annotations

import json
from pathlib import Path
import pytest

from evohomeasync2.schema import SCH_LOCN_STATUS
from evohomeasync2.schema.config import SCH_TEMPERATURE_CONTROL_SYSTEM, SCH_TIME_ZONE


TEST_DIR = Path(__file__).resolve().parent
WORK_DIR = f"{TEST_DIR}/schemas"


def pytest_generate_tests(metafunc: pytest.Metafunc):
    def id_fnc(folder_path: Path):
        return folder_path.name

    folders = [
        p for p in Path(WORK_DIR).glob("*") if p.is_dir() and not p.name.startswith("_")
    ]
    metafunc.parametrize("folder", sorted(folders), ids=id_fnc)


def _test_schema(folder: Path, schema: str, file_name: str):
    if not Path(folder).joinpath(file_name).is_file():
        pytest.skip(f"No {file_name} in: {folder.name}")

    with open(Path(folder).joinpath(file_name)) as f:
        data: dict = json.load(f)

    _ = schema(data)


# NOTE: JSON is not compliant with the schema, but data is useful to test against
def test_schemas_config(folder: Path):
    """Test the config schema for a location."""

    FILE_NAME = "config.json"

    if not Path(folder).joinpath(FILE_NAME).is_file():
        pytest.skip(f"No {FILE_NAME} in: {folder.name}")

    with open(Path(folder).joinpath(FILE_NAME)) as f:
        data: dict = json.load(f)

    _ = SCH_TIME_ZONE(data["locationInfo"]["timeZone"])
    for gwy_config in data["gateways"]:
        for tcs_config in gwy_config["temperatureControlSystems"]:
            _ = SCH_TEMPERATURE_CONTROL_SYSTEM(tcs_config)


def test_schemas_status(folder: Path):
    """Test the status schema for a location."""

    FILE_NAME = "status.json"

    if not Path(folder).joinpath(FILE_NAME).is_file():
        pytest.skip(f"No {FILE_NAME} in: {folder.name}")

    with open(Path(folder).joinpath(FILE_NAME)) as f:
        data: dict = json.load(f)

    _ = SCH_LOCN_STATUS(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohome-async - validate the config, status & schedule schemas."""
from __future__ import annotations

import asyncio
from datetime import datetime as dt
import logging
import os
from http import HTTPMethod, HTTPStatus

import evohomeasync2 as evo
from evohomeasync2.const import URL_BASE
from evohomeasync2.schema import vol  # voluptuous

from . import _DISABLE_STRICT_ASSERTS, _DEBUG_USE_REAL_AIOHTTP
from . import mocked_server as mock


if _DEBUG_USE_REAL_AIOHTTP:
    import aiohttp
else:
    from .mocked_server import aiohttp


_LOGGER = logging.getLogger(__name__)


def user_credentials():
    username = os.getenv("PYTEST_USERNAME")
    password = os.getenv("PYTEST_PASSWORD")

    # with open() as f:
    #     lines = f.readlines()
    # username = lines[0].strip()
    # password = lines[1].strip()

    return username, password


def client_session():
    if _DEBUG_USE_REAL_AIOHTTP:
        return aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=30))
    return aiohttp.ClientSession(mocked_server=mock.MockedServer(None, None))


def extract_oauth_tokens(client: evo.EvohomeClient):
    return (
        client.refresh_token,
        client.access_token,
        client.access_token_expires,
    )


async def should_work(
    client: evo.EvohomeClient,
    method: HTTPMethod,
    url: str,
    json: dict | None = None,
    content_type: str | None = "application/json",
    schema: vol.Schema | None = None,
) -> dict | list:
    """Make a request that is expected to succeed."""

    response: aiohttp.ClientResponse

    response, content = await client._broker._client(
        method, f"{URL_BASE}/{url}", json=json
    )

    response.raise_for_status()

    assert response.content_type == content_type

    if schema:
        return schema(content)
    return content


async def should_fail(
    client: evo.EvohomeClient,
    method: HTTPMethod,
    url: str,
    json: dict | None = None,
    content_type: str | None = "application/json",
    status: HTTPStatus | None = None,
) -> aiohttp.ClientResponse:
    """Make a request that is expected to fail."""

    response: aiohttp.ClientResponse

    response, content = await client._broker._client(
        method, f"{URL_BASE}/{url}", json=json
    )

    try:
        response.raise_for_status()
    except aiohttp.ClientResponseError as exc:
        assert exc.status == status
    else:
        assert False, response.status

    if _DISABLE_STRICT_ASSERTS:
        return response

    assert response.content_type == content_type

    if isinstance(content, list):
        assert status in (
            HTTPStatus.BAD_REQUEST,
            HTTPStatus.UNAUTHORIZED,
        ), response.status
        assert "message" in content[0]  # sometimes "code" too

    elif isinstance(content, dict):
        assert status in (
            HTTPStatus.NOT_FOUND,
            HTTPStatus.METHOD_NOT_ALLOWED,
        ), response.status
        assert "message" in content  # sometimes "code" too

    elif isinstance(content, str):  # 404
        assert status in (HTTPStatus.NOT_FOUND,), response.status

    else:
        assert False, response.status

    return response


async def wait_for_comm_task(
    client: evo.EvohomeClient,
    task_id: str,
    timeout: int = 3,
) -> None:
    """Wait for a communication task (API call) to complete."""

    await asyncio.sleep(0.5)

    url = f"commTasks?commTaskId={task_id}"

    start_time = dt.now()
    while True:
        response = await should_work(client, HTTPMethod.GET, url)
        if response["state"] == "Succeeded":
            return True
        if (dt.now() - start_time).total_seconds() > timeout:
            return False
        if response["state"] in ("Created", "Running"):
            await asyncio.sleep(0.3)
            continue
        else:
            # raise RuntimeError(f"Unexpected state: {response['state']}")
            _LOGGER.warning(f"Unexpected state: {response['state']}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 provides an async client for the *updated* Evohome API.

It is (largely) a faithful port of https://github.com/watchforstock/evohome-client

Further information at: https://evohome-client.readthedocs.io
"""

from .broker import Broker  # noqa: F401
from .client import EvohomeClient  # noqa: F401
from .controlsystem import ControlSystem  # noqa: F401
from .exceptions import (  # noqa: F401
    AuthenticationFailed,
    DeprecationError,
    EvohomeError,
    InvalidParameter,
    InvalidSchedule,
    InvalidSchema,
    RequestFailed,
    NoSingleTcsError,
    RateLimitExceeded,
)
from .gateway import Gateway  # noqa: F401
from .hotwater import HotWater  # noqa: F401
from .location import Location  # noqa: F401
from .zone import Zone  # noqa: F401


__version__ = "0.4.5"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Provides handling of TCC zones (heating and DHW)."""

# TODO: add provision for cooling zones
# TODO: add set_mode() for non-evohome modes (e.g. "VacationHold")

from __future__ import annotations

from datetime import datetime as dt
import json
from typing import TYPE_CHECKING, Final, NoReturn

from .const import API_STRFTIME, ZoneMode
from .exceptions import DeprecationError, InvalidSchedule, InvalidSchema
from .schema import SCH_ZONE_STATUS
from .schema.schedule import (
    SCH_GET_SCHEDULE,
    SCH_GET_SCHEDULE_ZONE,
    SCH_PUT_SCHEDULE,
    SCH_PUT_SCHEDULE_ZONE,
    convert_to_put_schedule,
)
from .schema.const import (
    SZ_ALLOWED_SETPOINT_MODES,
    SZ_ACTIVE_FAULTS,
    SZ_HEAT_SETPOINT_VALUE,
    SZ_MODEL_TYPE,
    SZ_NAME,
    SZ_SCHEDULE_CAPABILITIES,
    SZ_SETPOINT_CAPABILITIES,
    SZ_SETPOINT_MODE,
    SZ_SETPOINT_STATUS,
    SZ_TEMPERATURE_STATUS,
    SZ_TEMPERATURE_ZONE,
    SZ_TIME_UNTIL,
    SZ_ZONE_ID,
    SZ_ZONE_TYPE,
)

if TYPE_CHECKING:
    import logging

    from . import Broker, ControlSystem
    from .schema import _EvoDictT, _EvoListT, _ZoneIdT


class _ZoneBaseDeprecated:
    """Deprecated attributes and methods removed from the evohome-client namespace."""

    @property
    def zone_type(self) -> NoReturn:
        raise DeprecationError("ZoneBase.zone_type is deprecated, use .TYPE")

    async def schedule(self) -> NoReturn:
        raise DeprecationError(
            "_ZoneBase.schedule() is deprecrated, use .get_schedule()"
        )


class _ZoneBase(_ZoneBaseDeprecated):
    """Provide the base for temperatureZone / domesticHotWater Zones."""

    STATUS_SCHEMA = dict

    SCH_SCHEDULE_GET = SCH_GET_SCHEDULE
    SCH_SCHEDULE_PUT = SCH_PUT_SCHEDULE

    _id: str  # .zoneId or .dhwId
    TYPE: str  # "temperatureZone", "domesticHotWater"

    def __init__(self, tcs: ControlSystem) -> None:
        self.tcs = tcs

        self._broker: Broker = tcs._broker
        self._logger: logging.Logger = tcs._logger

        self._status: _EvoDictT = {}

    def __str__(self) -> str:
        return f"{self._id} ({self.TYPE})"

    async def _refresh_status(self) -> _EvoDictT:
        """Update the DHW/zone with its latest status (also returns the status).

        It will be more efficient to call Location.refresh_status().
        """

        self._logger.debug(f"Getting status of {self._id} ({self.TYPE})...")

        status: _EvoDictT = await self._broker.get(
            f"{self.TYPE}/{self._id}/status", schema=self.STATUS_SCHEMA
        )  # type: ignore[assignment]

        self._update_status(status)
        return status

    def _update_status(self, status: _EvoDictT) -> None:
        self._status = status

    @property
    def activeFaults(self) -> _EvoListT | None:
        return self._status.get(SZ_ACTIVE_FAULTS)

    @property
    def temperatureStatus(self) -> _EvoDictT | None:
        return self._status.get(SZ_TEMPERATURE_STATUS)

    async def get_schedule(self) -> _EvoDictT:
        """Get the schedule for this DHW/zone object."""

        self._logger.debug(f"Getting schedule of {self._id} ({self.TYPE})...")

        schedule: _EvoDictT = await self._broker.get(
            f"{self.TYPE}/{self._id}/schedule", schema=self.SCH_SCHEDULE_GET
        )  # type: ignore[assignment]

        return convert_to_put_schedule(schedule)

    async def set_schedule(self, schedule: _EvoDictT | str) -> None:
        """Set the schedule for this DHW/zone object."""

        self._logger.debug(f"Setting schedule of {self._id} ({self.TYPE})...")

        if isinstance(schedule, dict):
            try:
                json.dumps(schedule)
            except (OverflowError, TypeError, ValueError) as exc:
                raise InvalidSchedule(f"Invalid schedule: {exc}")

        elif isinstance(schedule, str):
            try:
                schedule = json.loads(schedule)
            except json.JSONDecodeError as exc:
                raise InvalidSchedule(f"Invalid schedule: {exc}")

        else:
            raise InvalidSchedule(f"Invalid schedule type: {type(schedule)}")

        _ = await self._broker.put(
            f"{self.TYPE}/{self._id}/schedule",
            json=schedule,
            schema=self.SCH_SCHEDULE_PUT,
        )


class _ZoneDeprecated:
    """Deprecated attributes and methods removed from the evohome-client namespace."""

    async def cancel_temp_override(self) -> None:
        raise DeprecationError(
            "Zone.cancel_temp_override() is deprecrated, use .reset_mode()"
        )


class Zone(_ZoneDeprecated, _ZoneBase):
    """Instance of a TCS's heating zone (temperatureZone)."""

    STATUS_SCHEMA = SCH_ZONE_STATUS
    TYPE: Final[str] = SZ_TEMPERATURE_ZONE  # type: ignore[misc]

    SCH_SCHEDULE_GET = SCH_GET_SCHEDULE_ZONE
    SCH_SCHEDULE_PUT = SCH_PUT_SCHEDULE_ZONE

    def __init__(self, tcs: ControlSystem, config: _EvoDictT) -> None:
        super().__init__(tcs)

        self._config: Final[_EvoDictT] = config

        try:
            assert self.zoneId, "Invalid config dict"
        except AssertionError as exc:
            raise InvalidSchema(str(exc))
        self._id = self.zoneId

    @property
    def zoneId(self) -> _ZoneIdT:
        return self._config[SZ_ZONE_ID]

    @property
    def modelType(self) -> str:
        return self._config[SZ_MODEL_TYPE]

    @property
    def setpointCapabilities(self) -> _EvoDictT:
        return self._config[SZ_SETPOINT_CAPABILITIES]

    @property  # for convenience (is not a top-level config attribute)
    def allowedSetpointModes(self) -> _EvoListT:
        return self._config[SZ_SETPOINT_CAPABILITIES][SZ_ALLOWED_SETPOINT_MODES]

    @property
    def scheduleCapabilities(self) -> _EvoDictT:
        return self._config[SZ_SCHEDULE_CAPABILITIES]

    @property
    def zoneType(self) -> str:
        return self._config[SZ_ZONE_TYPE]

    @property
    def name(self) -> str:
        return self._config.get(SZ_NAME) or self._config[SZ_NAME]

    @property
    def setpointStatus(self) -> _EvoDictT | None:
        return self._status.get(SZ_SETPOINT_STATUS)

    async def _set_mode(self, mode: dict) -> None:  # NOTE: no provision for cooling
        """Set the zone mode (heat_setpoint, cooling is TBD)."""
        _ = await self._broker.put(f"{self.TYPE}/{self._id}/heatSetpoint", json=mode)

    async def reset_mode(self) -> None:
        """Cancel any override and allow the zone to follow its schedule"""

        mode: dict[str, str | float | None] = {
            SZ_SETPOINT_MODE: ZoneMode.FOLLOW_SCHEDULE,
            SZ_HEAT_SETPOINT_VALUE: 0.0,
            SZ_TIME_UNTIL: None,
        }

        await self._set_mode(mode)

    async def set_temperature(  # NOTE: no provision for cooling
        self, temperature: float, /, *, until: dt | None = None
    ) -> None:
        """Set the temperature of the given zone (no provision for cooling)."""

        mode: dict[str, str | float | None]

        if until is None:  # NOTE: beware that these may be case-sensitive
            mode = {
                SZ_SETPOINT_MODE: ZoneMode.PERMANENT_OVERRIDE,
                SZ_HEAT_SETPOINT_VALUE: temperature,
                SZ_TIME_UNTIL: None,
            }
        else:
            mode = {
                SZ_SETPOINT_MODE: ZoneMode.TEMPORARY_OVERRIDE,
                SZ_HEAT_SETPOINT_VALUE: temperature,
                SZ_TIME_UNTIL: until.strftime(API_STRFTIME),
            }

        await self._set_mode(mode)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 - Schema for vendor's RESTful API JSON."""
from __future__ import annotations

try:
    from .account import SCH_USER_ACCOUNT as SCH_USER_ACCOUNT
    from .config import SCH_LOCATION_INSTALLATION_INFO as SCH_LOCN_CONFIG
    from .config import SCH_USER_LOCATIONS_INSTALLATION_INFO as SCH_FULL_CONFIG
    from .schedule import SCH_GET_SCHEDULE, SCH_PUT_SCHEDULE
    from .status import SCH_DHW as SCH_DHW_STATUS
    from .status import SCH_LOCATION_STATUS as SCH_LOCN_STATUS
    from .status import SCH_TEMPERATURE_CONTROL_SYSTEM as SCH_TCS_STATUS
    from .status import SCH_ZONE as SCH_ZONE_STATUS

except (AttributeError, ModuleNotFoundError):
    # No module named 'voluptuous'
    # type object 'vol' has no attribute 'Required'
    SCH_DHW_STATUS = dict
    SCH_FULL_CONFIG = list
    SCH_LOCN_CONFIG = dict
    SCH_LOCN_STATUS = dict
    SCH_OAUTH_TOKEN = dict
    SCH_TCS_STATUS = dict
    SCH_USER_ACCOUNT = dict
    SCH_GET_SCHEDULE = dict
    SCH_PUT_SCHEDULE = dict
    SCH_ZONE_STATUS = dict

from .const import (  # noqa: F401
    DHW_STATES,
    SYSTEM_MODES,
    ZONE_MODES,
    DhwState,
    SystemMode,
    ZoneMode,
)
from .helpers import vol  # voluptuous  # noqa: F401
from .schedule import convert_to_get_schedule  # noqa: F401
from .typing import _FilePathT, _ModeT  # noqa: F401
from .typing import (  # noqa: F401
    _DhwIdT,
    _GatewayIdT,
    _LocationIdT,
    _SystemIdT,
    _UserIdT,
    _ZoneIdT,
)
from .typing import _EvoLeafT, _EvoDictT, _EvoListT, _EvoSchemaT  # noqa: F401



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohome-async - validate the config & status schemas."""
from __future__ import annotations

import json
from pathlib import Path
import pytest

from evohomeasync2.schema import SCH_USER_ACCOUNT, SCH_FULL_CONFIG, SCH_LOCN_STATUS


TEST_DIR = Path(__file__).resolve().parent
WORK_DIR = f"{TEST_DIR}/systems"


def pytest_generate_tests(metafunc: pytest.Metafunc):
    def id_fnc(folder_path: Path):
        return folder_path.name

    folders = [
        p for p in Path(WORK_DIR).glob("*") if p.is_dir() and not p.name.startswith("_")
    ]
    metafunc.parametrize("folder", sorted(folders), ids=id_fnc)


def _test_schema(folder: Path, schema: str, file_name: str):
    if not Path(folder).joinpath(file_name).is_file():
        pytest.skip(f"No {file_name} in: {folder.name}")

    with open(Path(folder).joinpath(file_name)) as f:
        data: dict = json.load(f)

    _ = schema(data)


def test_user_account(folder: Path):
    """Test the user account schema against the corresponding JSON."""
    _test_schema(folder, SCH_USER_ACCOUNT, "user_account.json")


def test_user_locations(folder: Path):
    """Test the user locations config schema against the corresponding JSON."""
    _test_schema(folder, SCH_FULL_CONFIG, "user_locations.json")


def test_location_status(folder: Path):
    """Test the location status schema against the corresponding JSON."""
    for p in Path(folder).glob("status_*.json"):
        _test_schema(folder, SCH_LOCN_STATUS, p.name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohome-async - validate the handling of vendor APIs (URLs)."""
from __future__ import annotations


from datetime import datetime as dt, timedelta as td
from http import HTTPMethod, HTTPStatus
import pytest
import pytest_asyncio

import evohomeasync2 as evo
from evohomeasync2.const import API_STRFTIME, SystemMode, ZoneMode
from evohomeasync2.schema import (
    SCH_FULL_CONFIG,
    SCH_LOCN_STATUS,
    SCH_USER_ACCOUNT,
    SCH_GET_SCHEDULE,
    SCH_TCS_STATUS,
    SCH_ZONE_STATUS,
)
from evohomeasync2.schema.const import (
    SZ_DAILY_SCHEDULES,
    SZ_HEAT_SETPOINT,
    SZ_HEAT_SETPOINT_VALUE,
    SZ_IS_AVAILABLE,
    SZ_IS_PERMANENT,
    SZ_MODE,
    SZ_PERMANENT,
    SZ_SYSTEM_MODE,
    SZ_SETPOINT_MODE,
    SZ_SWITCHPOINTS,
    SZ_TEMPERATURE,
    SZ_TIME_UNTIL,
)
from evohomeasync2.schema.schedule import convert_to_put_schedule

from . import _DEBUG_USE_REAL_AIOHTTP
from .helpers import aiohttp, extract_oauth_tokens  # aiohttp may be mocked
from .helpers import user_credentials as _user_credentials
from .helpers import client_session as _client_session
from .helpers import should_work, should_fail


_global_oauth_tokens: tuple[str, str, dt] = None, None, None


@pytest.fixture()
def user_credentials():
    return _user_credentials()


@pytest_asyncio.fixture
async def session():
    client_session = _client_session()
    try:
        yield client_session
    finally:
        await client_session.close()


@pytest.fixture(autouse=True)
def patches_for_tests(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("evohomeasync2.broker.aiohttp", aiohttp)


async def instantiate_client(
    username: str,
    password: str,
    session: aiohttp.ClientSession | None = None,
):
    """Instantiate a client, and logon to the vendor API."""

    global _global_oauth_tokens

    refresh_token, access_token, access_token_expires = _global_oauth_tokens

    # Instantiation, NOTE: No API calls invoked during instantiation
    client = evo.EvohomeClient(
        username,
        password,
        session=session,
        refresh_token=refresh_token,
        access_token=access_token,
        access_token_expires=access_token_expires,
    )

    # Authentication
    await client._broker._basic_login()
    _global_oauth_tokens = extract_oauth_tokens(client)

    return client


async def _test_usr_account(
    username: str, password: str, session: aiohttp.ClientSession | None = None
) -> None:
    """Test /userAccount"""

    client = await instantiate_client(username, password, session=session)
    #

    url = "userAccount"
    _ = await should_work(client, HTTPMethod.GET, url, schema=SCH_USER_ACCOUNT)
    _ = await should_fail(
        client, HTTPMethod.PUT, url, status=HTTPStatus.METHOD_NOT_ALLOWED
    )

    url = "userXxxxxxx"  # NOTE: is a general test, and not a test specific to this URL
    _ = await should_fail(
        client,
        HTTPMethod.GET,
        url,
        status=HTTPStatus.NOT_FOUND,
        content_type="text/html",  # exception to usual content-type
    )


async def _test_all_config(
    username: str,
    password: str,
    session: aiohttp.ClientSession | None = None,
) -> None:
    """Test /location/installationInfo?userId={userId}"""

    client = await instantiate_client(username, password, session=session)
    _ = await client.user_account()
    #

    url = f"location/installationInfo?userId={client.account_info['userId']}"
    await should_work(client, HTTPMethod.GET, url)

    url += "&includeTemperatureControlSystems=True"
    _ = await should_work(client, HTTPMethod.GET, url, schema=SCH_FULL_CONFIG)
    _ = await should_fail(
        client, HTTPMethod.PUT, url, status=HTTPStatus.METHOD_NOT_ALLOWED
    )

    url = "location/installationInfo"
    _ = await should_fail(client, HTTPMethod.GET, url, status=HTTPStatus.NOT_FOUND)

    url = "location/installationInfo?userId=1230000"
    _ = await should_fail(client, HTTPMethod.GET, url, status=HTTPStatus.UNAUTHORIZED)

    url = "location/installationInfo?userId=xxxxxxx"
    _ = await should_fail(client, HTTPMethod.GET, url, status=HTTPStatus.BAD_REQUEST)

    url = "location/installationInfo?xxxxXx=xxxxxxx"
    _ = await should_fail(client, HTTPMethod.GET, url, status=HTTPStatus.NOT_FOUND)


async def _test_loc_status(
    username: str,
    password: str,
    session: aiohttp.ClientSession | None = None,
) -> None:
    """Test /location/{locationId}/status"""

    client = await instantiate_client(username, password, session=session)
    _ = await client.user_account()

    _ = await client._installation(refresh_status=False)
    loc = client.locations[0]
    #

    url = f"location/{loc.locationId}/status"
    _ = await should_work(client, HTTPMethod.GET, url)

    url += "?includeTemperatureControlSystems=True"
    _ = await should_work(client, HTTPMethod.GET, url, schema=SCH_LOCN_STATUS)
    _ = await should_fail(
        client, HTTPMethod.PUT, url, status=HTTPStatus.METHOD_NOT_ALLOWED
    )

    url = f"location/{loc.locationId}"
    await should_fail(
        client,
        HTTPMethod.GET,
        url,
        status=HTTPStatus.NOT_FOUND,
        content_type="text/html",  # exception to usual content-type
    )

    url = "location/1230000/status"
    _ = await should_fail(client, HTTPMethod.GET, url, status=HTTPStatus.UNAUTHORIZED)

    url = "location/xxxxxxx/status"
    _ = await should_fail(client, HTTPMethod.GET, url, status=HTTPStatus.BAD_REQUEST)

    url = f"location/{loc.locationId}/xxxxxxx"
    _ = await should_fail(
        client,
        HTTPMethod.GET,
        url,
        status=HTTPStatus.NOT_FOUND,
        content_type="text/html",  # exception to usual content-type
    )


async def _test_tcs_mode(
    username: str,
    password: str,
    session: aiohttp.ClientSession | None = None,
) -> None:
    """Test /temperatureControlSystem/{systemId}/mode"""

    client = await instantiate_client(username, password, session=session)
    _ = await client.user_account()
    _ = await client._installation(refresh_status=False)

    tcs: evo.ControlSystem

    if not (tcs := client.locations[0]._gateways[0]._control_systems[0]):
        pytest.skip("No available zones found")

    _ = await tcs._refresh_status()
    old_mode = tcs.systemModeStatus

    url = f"{tcs.TYPE}/{tcs._id}/status"
    _ = await should_work(client, HTTPMethod.GET, url, schema=SCH_TCS_STATUS)

    url = f"{tcs.TYPE}/{tcs._id}/mode"
    _ = await should_fail(
        client, HTTPMethod.GET, url, status=HTTPStatus.METHOD_NOT_ALLOWED
    )
    _ = await should_fail(
        client, HTTPMethod.PUT, url, json=old_mode, status=HTTPStatus.BAD_REQUEST
    )

    old_mode[SZ_SYSTEM_MODE] = old_mode.pop(SZ_MODE)
    old_mode[SZ_PERMANENT] = old_mode.pop(SZ_IS_PERMANENT)

    assert SystemMode.AUTO in [m[SZ_SYSTEM_MODE] for m in tcs.allowedSystemModes]
    new_mode = {
        SZ_SYSTEM_MODE: SystemMode.AUTO,
        SZ_PERMANENT: True,
    }
    _ = await should_work(client, HTTPMethod.PUT, url, json=new_mode)

    assert SystemMode.AWAY in [m[SZ_SYSTEM_MODE] for m in tcs.allowedSystemModes]
    new_mode = {
        SZ_SYSTEM_MODE: SystemMode.AWAY,
        SZ_PERMANENT: True,
        SZ_TIME_UNTIL: (dt.now() + td(hours=1)).strftime(API_STRFTIME),
    }
    _ = await should_work(client, HTTPMethod.PUT, url, json=new_mode)

    _ = await should_work(client, HTTPMethod.PUT, url, json=old_mode)

    url = f"{tcs.TYPE}/1234567/mode"
    _ = await should_fail(
        client, HTTPMethod.PUT, url, json=old_mode, status=HTTPStatus.UNAUTHORIZED
    )

    url = f"{tcs.TYPE}/{tcs._id}/systemMode"
    _ = await should_fail(
        client,
        HTTPMethod.PUT,
        url,
        json=old_mode,
        status=HTTPStatus.NOT_FOUND,
        content_type="text/html",  # exception to usual content-type
    )

    pass


async def _test_zone_mode(
    username: str,
    password: str,
    session: aiohttp.ClientSession | None = None,
) -> None:
    """Test /temperatureZone/{zoneId}/heatSetpoint"""

    client = await instantiate_client(username, password, session=session)
    _ = await client.user_account()
    _ = await client._installation(refresh_status=False)

    for zone in client.locations[0]._gateways[0]._control_systems[0]._zones:
        _ = await zone._refresh_status()
        if zone.temperatureStatus[SZ_IS_AVAILABLE]:
            break
    else:
        pytest.skip("No available zones found")
    #

    url = f"{zone.TYPE}/{zone._id}/status"
    _ = await should_work(client, HTTPMethod.GET, url, schema=SCH_ZONE_STATUS)

    url = f"{zone.TYPE}/{zone._id}/heatSetpoint"

    heat_setpoint = {
        SZ_SETPOINT_MODE: ZoneMode.PERMANENT_OVERRIDE,
        SZ_HEAT_SETPOINT_VALUE: zone.temperatureStatus[SZ_TEMPERATURE],
        # SZ_TIME_UNTIL: None,
    }
    _ = await should_work(client, HTTPMethod.PUT, url, json=heat_setpoint)

    heat_setpoint = {
        SZ_SETPOINT_MODE: ZoneMode.PERMANENT_OVERRIDE,
        SZ_HEAT_SETPOINT_VALUE: 99,
        SZ_TIME_UNTIL: None,
    }
    _ = await should_work(client, HTTPMethod.PUT, url, json=heat_setpoint)

    heat_setpoint = {
        SZ_SETPOINT_MODE: ZoneMode.PERMANENT_OVERRIDE,
        SZ_HEAT_SETPOINT_VALUE: 19.5,
    }
    _ = await should_work(client, HTTPMethod.PUT, url, json=heat_setpoint)

    heat_setpoint = {
        SZ_SETPOINT_MODE: "xxxxxxx",
        SZ_HEAT_SETPOINT_VALUE: 19.5,
    }
    _ = await should_fail(
        client, HTTPMethod.PUT, url, json=heat_setpoint, status=HTTPStatus.BAD_REQUEST
    )

    heat_setpoint = {
        SZ_SETPOINT_MODE: ZoneMode.FOLLOW_SCHEDULE,
        SZ_HEAT_SETPOINT_VALUE: 0.0,
        SZ_TIME_UNTIL: None,
    }
    _ = await should_work(client, HTTPMethod.PUT, url, json=heat_setpoint)


# TODO: Test sending bad schedule
# TODO: Try with/without convert_to_put_schedule()
async def _test_schedule(
    username: str,
    password: str,
    session: aiohttp.ClientSession | None = None,
) -> None:
    """Test /{x.TYPE}/{x._id}/schedule (of a zone)"""

    client = await instantiate_client(username, password, session=session)
    _ = await client.user_account()

    _ = await client.installation()
    zone = client.locations[0]._gateways[0]._control_systems[0]._zones[0]
    #

    url = f"{zone.TYPE}/{zone._id}/schedule"
    schedule = await should_work(client, HTTPMethod.GET, url, schema=SCH_GET_SCHEDULE)

    temp = schedule[SZ_DAILY_SCHEDULES][0][SZ_SWITCHPOINTS][0][SZ_HEAT_SETPOINT]

    schedule[SZ_DAILY_SCHEDULES][0][SZ_SWITCHPOINTS][0][SZ_HEAT_SETPOINT] = temp + 1
    _ = await should_work(
        client, HTTPMethod.PUT, url, json=convert_to_put_schedule(schedule)
    )

    schedule = await should_work(client, HTTPMethod.GET, url, schema=SCH_GET_SCHEDULE)
    assert (
        schedule[SZ_DAILY_SCHEDULES][0][SZ_SWITCHPOINTS][0][SZ_HEAT_SETPOINT]
        == temp + 1
    )

    schedule[SZ_DAILY_SCHEDULES][0][SZ_SWITCHPOINTS][0][SZ_HEAT_SETPOINT] = temp
    _ = await should_work(
        client, HTTPMethod.PUT, url, json=convert_to_put_schedule(schedule)
    )

    schedule = await should_work(client, HTTPMethod.GET, url, schema=SCH_GET_SCHEDULE)
    assert schedule[SZ_DAILY_SCHEDULES][0][SZ_SWITCHPOINTS][0][SZ_HEAT_SETPOINT] == temp

    _ = await should_fail(
        client, HTTPMethod.PUT, url, json=None, status=HTTPStatus.BAD_REQUEST
    )  # NOTE: json=None


@pytest.mark.asyncio
async def test_usr_account(
    user_credentials: tuple[str, str], session: aiohttp.ClientSession
) -> None:
    """Test /userAccount"""

    try:
        await _test_usr_account(*user_credentials, session=session)
    except evo.AuthenticationFailed:
        if not _DEBUG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Unable to authenticate")


@pytest.mark.asyncio
async def test_all_config(
    user_credentials: tuple[str, str], session: aiohttp.ClientSession
) -> None:
    """Test /location/installationInfo"""

    try:
        await _test_all_config(*user_credentials, session=session)
    except evo.AuthenticationFailed:
        if not _DEBUG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Unable to authenticate")


@pytest.mark.asyncio
async def test_loc_status(
    user_credentials: tuple[str, str], session: aiohttp.ClientSession
) -> None:
    """Test /location/{locationId}/status"""

    try:
        await _test_loc_status(*user_credentials, session=session)
    except evo.AuthenticationFailed:
        if not _DEBUG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Unable to authenticate")


@pytest.mark.asyncio
async def test_tcs_mode(
    user_credentials: tuple[str, str], session: aiohttp.ClientSession
) -> None:
    """Test /temperatureControlSystem/{systemId}/mode"""

    try:
        await _test_tcs_mode(*user_credentials, session=session)
    except evo.AuthenticationFailed:
        if not _DEBUG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Unable to authenticate")
    except NotImplementedError:  # TODO: implement
        if _DEBUG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Mocked server API not implemented")


@pytest.mark.asyncio
async def test_zone_mode(
    user_credentials: tuple[str, str], session: aiohttp.ClientSession
) -> None:
    """Test /temperatureZone/{zoneId}/heatSetpoint"""

    try:
        await _test_zone_mode(*user_credentials, session=session)
    except evo.AuthenticationFailed:
        if not _DEBUG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Unable to authenticate")
    except NotImplementedError:  # TODO: implement
        if _DEBUG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Mocked server API not implemented")


@pytest.mark.asyncio
async def test_schedule(
    user_credentials: tuple[str, str], session: aiohttp.ClientSession
) -> None:
    """Test /{x.TYPE}/{x_id}/schedule"""

    try:
        await _test_schedule(*user_credentials, session=session)
    except evo.AuthenticationFailed:
        if not _DEBUG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Unable to authenticate")


# TODO: test_oauth_token(
# TODO: test_put_dhw_state(
# TODO: test_get_dhw_status(
# TODO: test_set_tcs_mode(
# TODO: test_get_zon_status(



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 - Schema for RESTful API Account JSON."""
from __future__ import annotations

try:  # voluptuous is an optional module...
    import voluptuous as vol  # type: ignore[import-untyped]

except ModuleNotFoundError:  # No module named 'voluptuous'

    class vol:  # type: ignore[no-redef]
        class Invalid(Exception):
            pass

        Schema = dict | list


def camel_case(s: str) -> str:
    """Convert a PascalCase string to camelCase."""
    return s[:1].lower() + s[1:]


def pascal_case(s: str) -> str:
    """Convert a camelCase string to PascalCase."""
    return s[:1].upper() + s[1:]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 - Schema for RESTful API Config JSON."""
from __future__ import annotations

from .const import (
    REGEX_DHW_ID,
    REGEX_GATEWAY_ID,
    REGEX_LOCATION_ID,
    REGEX_SYSTEM_ID,
    REGEX_ZONE_ID,
    SZ_ACTIVE_FAULTS,
    SZ_DHW,
    SZ_DHW_ID,
    SZ_FAULT_TYPE,
    SZ_GATEWAY_ID,
    SZ_GATEWAYS,
    SZ_IS_AVAILABLE,
    SZ_IS_PERMANENT,
    SZ_LOCATION_ID,
    SZ_MODE,
    SZ_NAME,
    SZ_SETPOINT_MODE,
    SZ_SETPOINT_STATUS,
    SZ_SINCE,
    SZ_STATE,
    SZ_STATE_STATUS,
    SZ_SYSTEM_ID,
    SZ_SYSTEM_MODE_STATUS,
    SZ_TARGET_HEAT_TEMPERATURE,
    SZ_TEMPERATURE,
    SZ_TEMPERATURE_CONTROL_SYSTEMS,
    SZ_TEMPERATURE_STATUS,
    SZ_TIME_UNTIL,
    SZ_UNTIL,
    SZ_ZONE_ID,
    SZ_ZONES,
)
from .const import DhwState, FaultType, SystemMode, ZoneMode
from .helpers import vol  # voluptuous


SCH_ACTIVE_FAULT = vol.Schema(
    {
        vol.Required(SZ_FAULT_TYPE): vol.Any(*(m.value for m in FaultType)),
        vol.Required(SZ_SINCE): vol.Any(
            vol.Datetime(format="%Y-%m-%dT%H:%M:%S"),  # faults for zones
            vol.Datetime(format="%Y-%m-%dT%H:%M:%S.%f"),
            str,  # HACK: "2023-05-04T18:47:36.7727046"  faults for gateways
        ),
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_TEMPERATURE_STATUS = vol.Any(
    vol.Schema(
        {vol.Required(SZ_IS_AVAILABLE): False},
        extra=vol.PREVENT_EXTRA,
    ),
    vol.Schema(
        {vol.Required(SZ_IS_AVAILABLE): True, vol.Required(SZ_TEMPERATURE): float},
        extra=vol.PREVENT_EXTRA,
    ),
    extra=vol.PREVENT_EXTRA,
)

SCH_SETPOINT_STATUS = vol.Schema(
    {
        vol.Required(SZ_TARGET_HEAT_TEMPERATURE): float,
        vol.Required(SZ_SETPOINT_MODE): vol.Any(*(m.value for m in ZoneMode)),
        vol.Optional(SZ_UNTIL): vol.Datetime(format="%Y-%m-%dT%H:%M:%SZ"),
    },
    extra=vol.PREVENT_EXTRA,
)  # NOTE: SZ_UNTIL is present only for some modes

SCH_ZONE = vol.Schema(
    {
        vol.Required(SZ_ZONE_ID): vol.Match(REGEX_ZONE_ID),
        vol.Required(SZ_NAME): str,
        vol.Required(SZ_TEMPERATURE_STATUS): SCH_TEMPERATURE_STATUS,
        vol.Required(SZ_SETPOINT_STATUS): SCH_SETPOINT_STATUS,
        vol.Required(SZ_ACTIVE_FAULTS): [SCH_ACTIVE_FAULT],
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_STATE_STATUS = vol.Schema(
    {
        vol.Required(SZ_STATE): vol.Any(*(m.value for m in DhwState)),
        vol.Required(SZ_MODE): vol.Any(*(m.value for m in ZoneMode)),
        vol.Optional(SZ_UNTIL): vol.Datetime(format="%Y-%m-%dT%H:%M:%SZ"),
    },
    extra=vol.PREVENT_EXTRA,
)  # NOTE: SZ_UNTIL is present only for some modes

SCH_DHW = vol.Schema(
    {
        vol.Required(SZ_DHW_ID): vol.Match(REGEX_DHW_ID),
        vol.Required(SZ_TEMPERATURE_STATUS): SCH_TEMPERATURE_STATUS,
        vol.Required(SZ_STATE_STATUS): SCH_STATE_STATUS,
        vol.Required(SZ_ACTIVE_FAULTS): [SCH_ACTIVE_FAULT],
    },
    extra=vol.PREVENT_EXTRA,
)


SCH_SYSTEM_MODE_STATUS = vol.Any(
    vol.Schema(
        {
            vol.Required(SZ_MODE): vol.Any(*(m.value for m in SystemMode)),
            vol.Required(SZ_IS_PERMANENT): True,
        }
    ),
    vol.Schema(
        {
            vol.Required(SZ_MODE): vol.Any(
                str(SystemMode.AUTO_WITH_ECO),
                str(SystemMode.AWAY),
                str(SystemMode.CUSTOM),
                str(SystemMode.DAY_OFF),
            ),
            vol.Required(SZ_TIME_UNTIL): vol.Datetime(format="%Y-%m-%dT%H:%M:%SZ"),
            vol.Required(SZ_IS_PERMANENT): False,
        }
    ),
    extra=vol.PREVENT_EXTRA,
)

SCH_TEMPERATURE_CONTROL_SYSTEM = vol.Schema(
    {
        vol.Required(SZ_SYSTEM_ID): vol.Match(REGEX_SYSTEM_ID),
        vol.Required(SZ_SYSTEM_MODE_STATUS): SCH_SYSTEM_MODE_STATUS,
        vol.Required(SZ_ZONES): [SCH_ZONE],
        vol.Optional(SZ_DHW): SCH_DHW,
        vol.Required(SZ_ACTIVE_FAULTS): [SCH_ACTIVE_FAULT],
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_GATEWAY = vol.Schema(
    {
        vol.Required(SZ_GATEWAY_ID): vol.Match(REGEX_GATEWAY_ID),
        vol.Required(SZ_TEMPERATURE_CONTROL_SYSTEMS): [SCH_TEMPERATURE_CONTROL_SYSTEM],
        vol.Required(SZ_ACTIVE_FAULTS): [SCH_ACTIVE_FAULT],
    },
    extra=vol.PREVENT_EXTRA,
)

# location/{location_id}/status?includeTemperatureControlSystems=True
SCH_LOCATION_STATUS = vol.Schema(
    {
        vol.Required(SZ_LOCATION_ID): vol.Match(REGEX_LOCATION_ID),
        vol.Required(SZ_GATEWAYS): [SCH_GATEWAY],
    },
    extra=vol.PREVENT_EXTRA,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 - Schema for RESTful API Account JSON."""
from __future__ import annotations

from .const import (
    SZ_DAILY_SCHEDULES,
    SZ_DAY_OF_WEEK,
    SZ_DHW_STATE,
    SZ_HEAT_SETPOINT,
    SZ_OFF,
    SZ_ON,
    SZ_SWITCHPOINTS,
    SZ_TEMPERATURE,
    SZ_TIME_OF_DAY,
)
from .const import DAYS_OF_WEEK
from .helpers import vol, pascal_case  # voluptuous
from .typing import _EvoDictT, _EvoListT


#
# These are returned from vendor's API (GET)...
SCH_GET_SWITCHPOINT_DHW = vol.Schema(  # TODO: checkme
    {
        vol.Required(SZ_DHW_STATE): vol.Any(SZ_ON, SZ_OFF),
        vol.Required(SZ_TIME_OF_DAY): vol.Datetime(format="%H:%M:00"),
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_GET_SWITCHPOINT_ZONE = vol.Schema(
    {
        vol.Required(SZ_HEAT_SETPOINT): vol.All(float, vol.Range(min=5, max=35)),
        vol.Required(SZ_TIME_OF_DAY): vol.Datetime(format="%H:%M:00"),
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_GET_DAY_OF_WEEK_DHW = vol.Schema(
    {
        vol.Required(SZ_DAY_OF_WEEK): vol.Any(*DAYS_OF_WEEK),
        vol.Required(SZ_SWITCHPOINTS): [SCH_GET_SWITCHPOINT_DHW],
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_GET_DAY_OF_WEEK_ZONE = vol.Schema(
    {
        vol.Required(SZ_DAY_OF_WEEK): vol.Any(*DAYS_OF_WEEK),
        vol.Required(SZ_SWITCHPOINTS): [SCH_GET_SWITCHPOINT_ZONE],
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_GET_SCHEDULE_DHW = vol.Schema(
    {
        vol.Required(SZ_DAILY_SCHEDULES): [SCH_GET_DAY_OF_WEEK_DHW],
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_GET_SCHEDULE_ZONE = vol.Schema(
    {
        vol.Required(SZ_DAILY_SCHEDULES): [SCH_GET_DAY_OF_WEEK_ZONE],
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_GET_SCHEDULE = vol.Schema(  # PUT /{self.TYPE}/{self._id}/schedule
    vol.Any(SCH_GET_SCHEDULE_DHW, SCH_GET_SCHEDULE_ZONE),
    extra=vol.PREVENT_EXTRA,
)


#
# These are as to be provided to the vendor's API (PUT)...
# This is after modified by evohome-client (PUT), an evohome-client anachronism?
SCH_PUT_SWITCHPOINT_DHW = vol.Schema(  # TODO: checkme
    {
        vol.Required(pascal_case(SZ_DHW_STATE)): vol.Any(SZ_ON, SZ_OFF),
        vol.Required(pascal_case(SZ_TIME_OF_DAY)): vol.Datetime(format="%H:%M:00"),
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_PUT_SWITCHPOINT_ZONE = vol.Schema(
    {  # NOTE: SZ_HEAT_SETPOINT is not .capitalized()
        vol.Required(SZ_HEAT_SETPOINT): vol.All(float, vol.Range(min=5, max=35)),
        vol.Required(pascal_case(SZ_TIME_OF_DAY)): vol.Datetime(format="%H:%M:00"),
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_PUT_DAY_OF_WEEK_DHW = vol.Schema(
    {
        vol.Required(pascal_case(SZ_DAY_OF_WEEK)): vol.Any(
            vol.All(int, vol.Range(min=0, max=6)),  # 0 is Monday
        ),
        vol.Required(pascal_case(SZ_SWITCHPOINTS)): [SCH_PUT_SWITCHPOINT_DHW],
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_PUT_DAY_OF_WEEK_ZONE = vol.Schema(
    {
        vol.Required(pascal_case(SZ_DAY_OF_WEEK)): vol.Any(
            vol.All(int, vol.Range(min=0, max=6)),  # 0 is Monday
        ),
        vol.Required(pascal_case(SZ_SWITCHPOINTS)): [SCH_PUT_SWITCHPOINT_ZONE],
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_PUT_SCHEDULE_DHW = vol.Schema(
    {
        vol.Required(pascal_case(SZ_DAILY_SCHEDULES)): [SCH_PUT_DAY_OF_WEEK_DHW],
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_PUT_SCHEDULE_ZONE = vol.Schema(
    {
        vol.Required(pascal_case(SZ_DAILY_SCHEDULES)): [SCH_PUT_DAY_OF_WEEK_ZONE],
    },
    extra=vol.PREVENT_EXTRA,
)

SCH_PUT_SCHEDULE = vol.Schema(  # PUT /{self.TYPE}/{self._id}/schedule
    vol.Any(SCH_PUT_SCHEDULE_DHW, SCH_PUT_SCHEDULE_ZONE),
    extra=vol.PREVENT_EXTRA,
)


#
#
def _convert_to_put_schedule_classic(raw_schedule: _EvoDictT) -> _EvoDictT:
    """Convert a schedule as per evohomeclient2.schedule()."""

    import json

    MAPPINGS = [
        (SZ_DAILY_SCHEDULES, "DailySchedules"),
        (SZ_DAY_OF_WEEK, "DayOfWeek"),
        (SZ_DHW_STATE, "DhwState"),
        (SZ_SWITCHPOINTS, "Switchpoints"),
        (SZ_TEMPERATURE, "TargetTemperature"),  # should have been SZ_HEAT_SETPOINT?
        (SZ_TIME_OF_DAY, "TimeOfDay"),
    ]

    response_data = json.dumps(raw_schedule)

    for from_key, to_key in MAPPINGS:  # an anachronism from evohome-client
        response_data = response_data.replace(from_key, to_key)

    data: _EvoDictT = json.loads(response_data)

    # change the day name string to an ordinal (Monday = 0)
    for day_of_week, setpoints in enumerate(data[pascal_case(SZ_DAILY_SCHEDULES)]):
        setpoints[pascal_case(SZ_DAY_OF_WEEK)] = day_of_week

    return data


def convert_to_put_schedule(get_schedule: _EvoDictT) -> _EvoDictT:
    """Convert a schedule to the format used by our get/set_schedule() methods.

    The 'raw' schedule format is the one returned by the vendor's RESTful API (GET).
    """

    put_schedule: dict[str, _EvoListT] = {}
    put_schedule[pascal_case(SZ_DAILY_SCHEDULES)] = []

    for day_of_week, get_schedule in enumerate(get_schedule[SZ_DAILY_SCHEDULES]):
        put_day_schedule: _EvoDictT = {pascal_case(SZ_DAY_OF_WEEK): day_of_week}
        put_switchpoints: _EvoListT = []

        for get_sp in get_schedule[SZ_SWITCHPOINTS]:
            if SZ_HEAT_SETPOINT in get_sp:
                # NOTE: this key is not converted to pascal_case in evohomeclient2
                put_sp = {SZ_HEAT_SETPOINT: get_sp[SZ_HEAT_SETPOINT]}
            else:
                put_sp = {pascal_case(SZ_DHW_STATE): get_sp[SZ_DHW_STATE]}

            put_sp[pascal_case(SZ_TIME_OF_DAY)] = get_sp[SZ_TIME_OF_DAY]
            put_switchpoints.append(put_sp)

        put_day_schedule[pascal_case(SZ_SWITCHPOINTS)] = put_switchpoints
        put_schedule[pascal_case(SZ_DAILY_SCHEDULES)].append(put_day_schedule)

    return put_schedule


def convert_to_get_schedule(put_schedule: _EvoDictT) -> _EvoDictT:
    """Convert a schedule to the format returned by the vendor's RESTful API (GET)."""

    get_schedule: dict[str, _EvoListT] = {}
    get_schedule[SZ_DAILY_SCHEDULES] = []

    for put_day_schedule in put_schedule[pascal_case(SZ_DAILY_SCHEDULES)]:
        day_of_week = put_day_schedule[pascal_case(SZ_DAY_OF_WEEK)]
        get_day_schedule: _EvoDictT = {SZ_DAY_OF_WEEK: DAYS_OF_WEEK[day_of_week]}
        get_switchpoints: _EvoListT = []

        for put_sp in put_day_schedule[pascal_case(SZ_SWITCHPOINTS)]:
            if SZ_HEAT_SETPOINT in put_sp:
                # NOTE: this key is not converted to pascal_case in evohomeclient2
                get_sp = {SZ_HEAT_SETPOINT: put_sp[SZ_HEAT_SETPOINT]}
            else:
                get_sp = {SZ_DHW_STATE: put_sp[pascal_case(SZ_DHW_STATE)]}

            get_sp[SZ_TIME_OF_DAY] = put_sp[pascal_case(SZ_TIME_OF_DAY)]
            get_switchpoints.append(get_sp)

        get_day_schedule[SZ_SWITCHPOINTS] = get_switchpoints
        get_schedule[SZ_DAILY_SCHEDULES].append(get_day_schedule)

    return get_schedule



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Mocked vendor RESTful API via a hacked aiohttp."""
from __future__ import annotations

from typing import Literal


# Sample responses from evohome-client
MOCK_AUTH_RESPONSE = {  # can use this for all
    "access_token": "ncWMqPh2yGgAqc...",
    "token_type": "bearer",
    "expires_in": 1800,
    "refresh_token": "Ryx9fL34Z5GcNV...",
    "scope": "EMEA-V1-Basic EMEA-V1-Anonymous",
}


MOCK_FULL_CONFIG = [  # of a specific user account
    {
        "locationInfo": {
            "locationId": "2738909",
            "name": "My Home",
            "streetAddress": "1 Main Street",
            "city": "London",
            "country": "UnitedKingdom",
            "postcode": "E1 1AA",
            "locationType": "Residential",
            "useDaylightSaveSwitching": True,
            "timeZone": {
                "timeZoneId": "GMTStandardTime",
                "displayName": "(UTC+00:00) Dublin, Edinburgh, Lisbon, London",
                "offsetMinutes": 0,
                "currentOffsetMinutes": 60,
                "supportsDaylightSaving": True,
            },
            "locationOwner": {
                "userId": "2263181",
                "username": "john.smith@gmail.com",
                "firstname": "John",
                "lastname": "Smith",
            },
        },
        "gateways": [
            {
                "gatewayInfo": {
                    "gatewayId": "2499896",
                    "mac": "00D02DEE1234",
                    "crc": "9999",
                    "isWiFi": False,
                },
                "temperatureControlSystems": [
                    {
                        "systemId": "3432522",
                        "modelType": "EvoTouch",
                        "zones": [
                            {
                                "zoneId": "3432521",
                                "modelType": "HeatingZone",
                                "setpointCapabilities": {
                                    "maxHeatSetpoint": 35.0,
                                    "minHeatSetpoint": 5.0,
                                    "valueResolution": 0.5,
                                    "canControlHeat": True,
                                    "canControlCool": False,
                                    "allowedSetpointModes": [
                                        "PermanentOverride",
                                        "FollowSchedule",
                                        "TemporaryOverride",
                                    ],
                                    "maxDuration": "1.00:00:00",
                                    "timingResolution": "00:10:00",
                                },
                                "scheduleCapabilities": {
                                    "maxSwitchpointsPerDay": 6,
                                    "minSwitchpointsPerDay": 1,
                                    "timingResolution": "00:10:00",
                                    "setpointValueResolution": 0.5,
                                },
                                "name": "Dead Zone",
                                "zoneType": "RadiatorZone",
                            },
                            {
                                "zoneId": "3432576",
                                "modelType": "HeatingZone",
                                "setpointCapabilities": {
                                    "maxHeatSetpoint": 35.0,
                                    "minHeatSetpoint": 5.0,
                                    "valueResolution": 0.5,
                                    "canControlHeat": True,
                                    "canControlCool": False,
                                    "allowedSetpointModes": [
                                        "PermanentOverride",
                                        "FollowSchedule",
                                        "TemporaryOverride",
                                    ],
                                    "maxDuration": "1.00:00:00",
                                    "timingResolution": "00:10:00",
                                },
                                "scheduleCapabilities": {
                                    "maxSwitchpointsPerDay": 6,
                                    "minSwitchpointsPerDay": 1,
                                    "timingResolution": "00:10:00",
                                    "setpointValueResolution": 0.5,
                                },
                                "name": "Main Room",
                                "zoneType": "RadiatorZone",
                            },
                            {
                                "zoneId": "3432577",
                                "modelType": "HeatingZone",
                                "setpointCapabilities": {
                                    "maxHeatSetpoint": 35.0,
                                    "minHeatSetpoint": 5.0,
                                    "valueResolution": 0.5,
                                    "canControlHeat": True,
                                    "canControlCool": False,
                                    "allowedSetpointModes": [
                                        "PermanentOverride",
                                        "FollowSchedule",
                                        "TemporaryOverride",
                                    ],
                                    "maxDuration": "1.00:00:00",
                                    "timingResolution": "00:10:00",
                                },
                                "scheduleCapabilities": {
                                    "maxSwitchpointsPerDay": 6,
                                    "minSwitchpointsPerDay": 1,
                                    "timingResolution": "00:10:00",
                                    "setpointValueResolution": 0.5,
                                },
                                "name": "Front Room",
                                "zoneType": "RadiatorZone",
                            },
                            {
                                "zoneId": "3432578",
                                "modelType": "HeatingZone",
                                "setpointCapabilities": {
                                    "maxHeatSetpoint": 35.0,
                                    "minHeatSetpoint": 5.0,
                                    "valueResolution": 0.5,
                                    "canControlHeat": True,
                                    "canControlCool": False,
                                    "allowedSetpointModes": [
                                        "PermanentOverride",
                                        "FollowSchedule",
                                        "TemporaryOverride",
                                    ],
                                    "maxDuration": "1.00:00:00",
                                    "timingResolution": "00:10:00",
                                },
                                "scheduleCapabilities": {
                                    "maxSwitchpointsPerDay": 6,
                                    "minSwitchpointsPerDay": 1,
                                    "timingResolution": "00:10:00",
                                    "setpointValueResolution": 0.5,
                                },
                                "name": "Kitchen",
                                "zoneType": "RadiatorZone",
                            },
                            {
                                "zoneId": "3432579",
                                "modelType": "HeatingZone",
                                "setpointCapabilities": {
                                    "maxHeatSetpoint": 35.0,
                                    "minHeatSetpoint": 5.0,
                                    "valueResolution": 0.5,
                                    "canControlHeat": True,
                                    "canControlCool": False,
                                    "allowedSetpointModes": [
                                        "PermanentOverride",
                                        "FollowSchedule",
                                        "TemporaryOverride",
                                    ],
                                    "maxDuration": "1.00:00:00",
                                    "timingResolution": "00:10:00",
                                },
                                "scheduleCapabilities": {
                                    "maxSwitchpointsPerDay": 6,
                                    "minSwitchpointsPerDay": 1,
                                    "timingResolution": "00:10:00",
                                    "setpointValueResolution": 0.5,
                                },
                                "name": "Bathroom Dn",
                                "zoneType": "RadiatorZone",
                            },
                            {
                                "zoneId": "3432580",
                                "modelType": "HeatingZone",
                                "setpointCapabilities": {
                                    "maxHeatSetpoint": 35.0,
                                    "minHeatSetpoint": 5.0,
                                    "valueResolution": 0.5,
                                    "canControlHeat": True,
                                    "canControlCool": False,
                                    "allowedSetpointModes": [
                                        "PermanentOverride",
                                        "FollowSchedule",
                                        "TemporaryOverride",
                                    ],
                                    "maxDuration": "1.00:00:00",
                                    "timingResolution": "00:10:00",
                                },
                                "scheduleCapabilities": {
                                    "maxSwitchpointsPerDay": 6,
                                    "minSwitchpointsPerDay": 1,
                                    "timingResolution": "00:10:00",
                                    "setpointValueResolution": 0.5,
                                },
                                "name": "Main Bedroom",
                                "zoneType": "RadiatorZone",
                            },
                            {
                                "zoneId": "3449703",
                                "modelType": "HeatingZone",
                                "setpointCapabilities": {
                                    "maxHeatSetpoint": 35.0,
                                    "minHeatSetpoint": 5.0,
                                    "valueResolution": 0.5,
                                    "canControlHeat": True,
                                    "canControlCool": False,
                                    "allowedSetpointModes": [
                                        "PermanentOverride",
                                        "FollowSchedule",
                                        "TemporaryOverride",
                                    ],
                                    "maxDuration": "1.00:00:00",
                                    "timingResolution": "00:10:00",
                                },
                                "scheduleCapabilities": {
                                    "maxSwitchpointsPerDay": 6,
                                    "minSwitchpointsPerDay": 1,
                                    "timingResolution": "00:10:00",
                                    "setpointValueResolution": 0.5,
                                },
                                "name": "Spare Room",
                                "zoneType": "RadiatorZone",
                            },
                            {
                                "zoneId": "3449740",
                                "modelType": "HeatingZone",
                                "setpointCapabilities": {
                                    "maxHeatSetpoint": 35.0,
                                    "minHeatSetpoint": 5.0,
                                    "valueResolution": 0.5,
                                    "canControlHeat": True,
                                    "canControlCool": False,
                                    "allowedSetpointModes": [
                                        "PermanentOverride",
                                        "FollowSchedule",
                                        "TemporaryOverride",
                                    ],
                                    "maxDuration": "1.00:00:00",
                                    "timingResolution": "00:10:00",
                                },
                                "scheduleCapabilities": {
                                    "maxSwitchpointsPerDay": 6,
                                    "minSwitchpointsPerDay": 1,
                                    "timingResolution": "00:10:00",
                                    "setpointValueResolution": 0.5,
                                },
                                "name": "Bathroom Up",
                                "zoneType": "RadiatorZone",
                            },
                            {
                                "zoneId": "3450733",
                                "modelType": "HeatingZone",
                                "setpointCapabilities": {
                                    "maxHeatSetpoint": 35.0,
                                    "minHeatSetpoint": 5.0,
                                    "valueResolution": 0.5,
                                    "canControlHeat": True,
                                    "canControlCool": False,
                                    "allowedSetpointModes": [
                                        "PermanentOverride",
                                        "FollowSchedule",
                                        "TemporaryOverride",
                                    ],
                                    "maxDuration": "1.00:00:00",
                                    "timingResolution": "00:10:00",
                                },
                                "scheduleCapabilities": {
                                    "maxSwitchpointsPerDay": 6,
                                    "minSwitchpointsPerDay": 1,
                                    "timingResolution": "00:10:00",
                                    "setpointValueResolution": 0.5,
                                },
                                "name": "Kids Room",
                                "zoneType": "RadiatorZone",
                            },
                        ],
                        "allowedSystemModes": [
                            {
                                "systemMode": "Auto",
                                "canBePermanent": True,
                                "canBeTemporary": False,
                            },
                            {
                                "systemMode": "AutoWithEco",
                                "canBePermanent": True,
                                "canBeTemporary": True,
                                "maxDuration": "1.00:00:00",
                                "timingResolution": "01:00:00",
                                "timingMode": "Duration",
                            },
                            {
                                "systemMode": "AutoWithReset",
                                "canBePermanent": True,
                                "canBeTemporary": False,
                            },
                            {
                                "systemMode": "Away",
                                "canBePermanent": True,
                                "canBeTemporary": True,
                                "maxDuration": "99.00:00:00",
                                "timingResolution": "1.00:00:00",
                                "timingMode": "Period",
                            },
                            {
                                "systemMode": "DayOff",
                                "canBePermanent": True,
                                "canBeTemporary": True,
                                "maxDuration": "99.00:00:00",
                                "timingResolution": "1.00:00:00",
                                "timingMode": "Period",
                            },
                            {
                                "systemMode": "HeatingOff",
                                "canBePermanent": True,
                                "canBeTemporary": False,
                            },
                            {
                                "systemMode": "Custom",
                                "canBePermanent": True,
                                "canBeTemporary": True,
                                "maxDuration": "99.00:00:00",
                                "timingResolution": "1.00:00:00",
                                "timingMode": "Period",
                            },
                        ],
                    }
                ],
            }
        ],
    }
]


MOCK_LOCN_STATUS = {  # of a specific location
    "locationId": "2738909",
    "gateways": [
        {
            "gatewayId": "2499896",
            "temperatureControlSystems": [
                {
                    "systemId": "3432522",
                    "zones": [
                        {
                            "zoneId": "3432521",
                            "temperatureStatus": {"isAvailable": False},
                            "activeFaults": [
                                {
                                    "faultType": "TempZoneSensorCommunicationLost",
                                    "since": "2023-10-09T01:45:00",
                                }
                            ],
                            "setpointStatus": {
                                "targetHeatTemperature": 5.0,
                                "setpointMode": "PermanentOverride",
                            },
                            "name": "Dead Zone",
                        },
                        {
                            "zoneId": "3432576",
                            "temperatureStatus": {
                                "temperature": 19.5,
                                "isAvailable": True,
                            },
                            "activeFaults": [],
                            "setpointStatus": {
                                "targetHeatTemperature": 5.0,
                                "setpointMode": "FollowSchedule",
                            },
                            "name": "Main Room",
                        },
                        {
                            "zoneId": "3432577",
                            "temperatureStatus": {
                                "temperature": 18.0,
                                "isAvailable": True,
                            },
                            "activeFaults": [],
                            "setpointStatus": {
                                "targetHeatTemperature": 5.0,
                                "setpointMode": "PermanentOverride",
                            },
                            "name": "Front Room",
                        },
                        {
                            "zoneId": "3432578",
                            "temperatureStatus": {"isAvailable": False},
                            "activeFaults": [
                                {
                                    "faultType": "TempZoneSensorLowBattery",
                                    "since": "2023-06-03T21:05:32",
                                },
                                {
                                    "faultType": "TempZoneSensorCommunicationLost",
                                    "since": "2023-09-02T12:34:46",
                                },
                            ],
                            "setpointStatus": {
                                "targetHeatTemperature": 5.0,
                                "setpointMode": "FollowSchedule",
                            },
                            "name": "Kitchen",
                        },
                        {
                            "zoneId": "3432579",
                            "temperatureStatus": {"isAvailable": False},
                            "activeFaults": [
                                {
                                    "faultType": "TempZoneSensorCommunicationLost",
                                    "since": "2023-10-09T01:45:04",
                                }
                            ],
                            "setpointStatus": {
                                "targetHeatTemperature": 5.0,
                                "setpointMode": "FollowSchedule",
                            },
                            "name": "Bathroom Dn",
                        },
                        {
                            "zoneId": "3432580",
                            "temperatureStatus": {
                                "temperature": 18.0,
                                "isAvailable": True,
                            },
                            "activeFaults": [],
                            "setpointStatus": {
                                "targetHeatTemperature": 5.0,
                                "setpointMode": "PermanentOverride",
                            },
                            "name": "Main Bedroom",
                        },
                        {
                            "zoneId": "3449703",
                            "temperatureStatus": {
                                "temperature": 18.0,
                                "isAvailable": True,
                            },
                            "activeFaults": [],
                            "setpointStatus": {
                                "targetHeatTemperature": 5.0,
                                "setpointMode": "PermanentOverride",
                            },
                            "name": "Spare Room",
                        },
                        {
                            "zoneId": "3449740",
                            "temperatureStatus": {
                                "temperature": 18.0,
                                "isAvailable": True,
                            },
                            "activeFaults": [],
                            "setpointStatus": {
                                "targetHeatTemperature": 5.0,
                                "setpointMode": "FollowSchedule",
                            },
                            "name": "Bathroom Up",
                        },
                        {
                            "zoneId": "3450733",
                            "temperatureStatus": {
                                "temperature": 18.5,
                                "isAvailable": True,
                            },
                            "activeFaults": [
                                {
                                    "faultType": "TempZoneActuatorCommunicationLost",
                                    "since": "2023-09-18T22:54:17",
                                }
                            ],
                            "setpointStatus": {
                                "targetHeatTemperature": 5.0,
                                "setpointMode": "PermanentOverride",
                            },
                            "name": "Kids Room",
                        },
                    ],
                    "activeFaults": [],
                    "systemModeStatus": {"mode": "HeatingOff", "isPermanent": True},
                }
            ],
            "activeFaults": [],
        }
    ],
}


MOCK_SCHEDULE_ZONE = {  # of any zone (no zoneId)
    "dailySchedules": [
        {
            "dayOfWeek": "Monday",
            "switchpoints": [
                {"heatSetpoint": 19.0000, "timeOfDay": "06:30:00"},
                {"heatSetpoint": 18.0000, "timeOfDay": "08:00:00"},
                {"heatSetpoint": 18.5000, "timeOfDay": "17:00:00"},
                {"heatSetpoint": 14.9000, "timeOfDay": "21:30:00"},
            ],
        },
        {
            "dayOfWeek": "Tuesday",
            "switchpoints": [
                {"heatSetpoint": 19.0000, "timeOfDay": "06:30:00"},
                {"heatSetpoint": 18.0000, "timeOfDay": "08:00:00"},
                {"heatSetpoint": 18.5000, "timeOfDay": "17:00:00"},
                {"heatSetpoint": 14.9000, "timeOfDay": "21:30:00"},
            ],
        },
        {
            "dayOfWeek": "Wednesday",
            "switchpoints": [
                {"heatSetpoint": 19.0000, "timeOfDay": "06:30:00"},
                {"heatSetpoint": 18.0000, "timeOfDay": "08:00:00"},
                {"heatSetpoint": 18.5000, "timeOfDay": "17:00:00"},
                {"heatSetpoint": 14.9000, "timeOfDay": "21:30:00"},
            ],
        },
        {
            "dayOfWeek": "Thursday",
            "switchpoints": [
                {"heatSetpoint": 19.0000, "timeOfDay": "06:30:00"},
                {"heatSetpoint": 18.0000, "timeOfDay": "08:00:00"},
                {"heatSetpoint": 18.5000, "timeOfDay": "17:00:00"},
                {"heatSetpoint": 14.9000, "timeOfDay": "21:30:00"},
            ],
        },
        {
            "dayOfWeek": "Friday",
            "switchpoints": [
                {"heatSetpoint": 19.0000, "timeOfDay": "06:30:00"},
                {"heatSetpoint": 18.0000, "timeOfDay": "08:00:00"},
                {"heatSetpoint": 18.5000, "timeOfDay": "17:00:00"},
                {"heatSetpoint": 14.9000, "timeOfDay": "21:30:00"},
            ],
        },
        {
            "dayOfWeek": "Saturday",
            "switchpoints": [
                {"heatSetpoint": 19.0000, "timeOfDay": "07:30:00"},
                {"heatSetpoint": 18.5000, "timeOfDay": "10:00:00"},
                {"heatSetpoint": 18.5000, "timeOfDay": "17:00:00"},
                {"heatSetpoint": 14.9000, "timeOfDay": "21:30:00"},
            ],
        },
        {
            "dayOfWeek": "Sunday",
            "switchpoints": [
                {"heatSetpoint": 19.0000, "timeOfDay": "07:30:00"},
                {"heatSetpoint": 18.5000, "timeOfDay": "10:00:00"},
                {"heatSetpoint": 18.5000, "timeOfDay": "17:00:00"},
                {"heatSetpoint": 14.9000, "timeOfDay": "21:30:00"},
            ],
        },
    ]
}


MOCK_SCHEDULE_DHW = {  # of any zone (no dhwId)
    "dailySchedules": [
        {
            "dayOfWeek": "Monday",
            "switchpoints": [
                {"dhwState": "On", "timeOfDay": "06:30:00"},
                {"dhwState": "Off", "timeOfDay": "08:30:00"},
                {"dhwState": "On", "timeOfDay": "12:00:00"},
                {"dhwState": "Off", "timeOfDay": "13:00:00"},
                {"dhwState": "On", "timeOfDay": "16:30:00"},
                {"dhwState": "Off", "timeOfDay": "22:30:00"},
            ],
        },
        {
            "dayOfWeek": "Tuesday",
            "switchpoints": [
                {"dhwState": "On", "timeOfDay": "06:30:00"},
                {"dhwState": "Off", "timeOfDay": "08:30:00"},
                {"dhwState": "On", "timeOfDay": "12:00:00"},
                {"dhwState": "Off", "timeOfDay": "13:00:00"},
                {"dhwState": "On", "timeOfDay": "16:30:00"},
                {"dhwState": "Off", "timeOfDay": "22:30:00"},
            ],
        },
        {
            "dayOfWeek": "Wednesday",
            "switchpoints": [
                {"dhwState": "On", "timeOfDay": "06:30:00"},
                {"dhwState": "Off", "timeOfDay": "08:30:00"},
                {"dhwState": "On", "timeOfDay": "12:00:00"},
                {"dhwState": "Off", "timeOfDay": "13:00:00"},
                {"dhwState": "On", "timeOfDay": "16:30:00"},
                {"dhwState": "Off", "timeOfDay": "22:30:00"},
            ],
        },
        {
            "dayOfWeek": "Thursday",
            "switchpoints": [
                {"dhwState": "On", "timeOfDay": "06:30:00"},
                {"dhwState": "Off", "timeOfDay": "08:30:00"},
                {"dhwState": "On", "timeOfDay": "12:00:00"},
                {"dhwState": "Off", "timeOfDay": "13:00:00"},
                {"dhwState": "On", "timeOfDay": "16:30:00"},
                {"dhwState": "Off", "timeOfDay": "22:30:00"},
            ],
        },
        {
            "dayOfWeek": "Friday",
            "switchpoints": [
                {"dhwState": "On", "timeOfDay": "06:30:00"},
                {"dhwState": "Off", "timeOfDay": "08:30:00"},
                {"dhwState": "On", "timeOfDay": "12:00:00"},
                {"dhwState": "Off", "timeOfDay": "13:00:00"},
                {"dhwState": "On", "timeOfDay": "16:30:00"},
                {"dhwState": "Off", "timeOfDay": "22:30:00"},
            ],
        },
        {
            "dayOfWeek": "Saturday",
            "switchpoints": [
                {"dhwState": "On", "timeOfDay": "06:30:00"},
                {"dhwState": "Off", "timeOfDay": "09:30:00"},
                {"dhwState": "On", "timeOfDay": "12:00:00"},
                {"dhwState": "Off", "timeOfDay": "13:00:00"},
                {"dhwState": "On", "timeOfDay": "16:30:00"},
                {"dhwState": "Off", "timeOfDay": "23:00:00"},
            ],
        },
        {
            "dayOfWeek": "Sunday",
            "switchpoints": [
                {"dhwState": "On", "timeOfDay": "06:30:00"},
                {"dhwState": "Off", "timeOfDay": "09:30:00"},
                {"dhwState": "On", "timeOfDay": "12:00:00"},
                {"dhwState": "Off", "timeOfDay": "13:00:00"},
                {"dhwState": "On", "timeOfDay": "16:30:00"},
                {"dhwState": "Off", "timeOfDay": "23:00:00"},
            ],
        },
    ]
}


def user_config_from_full_config(full_config: list) -> dict:
    """Create a valid MOCK_USER_CONFIG from a MOCK_FULL_CONFIG."""

    # assert schema
    loc_idx = 0
    return (
        full_config[loc_idx]["locationInfo"]["locationOwner"]
        | {
            k: v
            for k, v in full_config[loc_idx]["locationInfo"].items()
            if k in ("streetAddress", "city", "postcode", "country")
        }
        | {"language": "enGB"}
    )


MOCK_USER_CONFIG = user_config_from_full_config(MOCK_FULL_CONFIG)


_bodyT = dict | str
_methodT = Literal["GET", "POST", "PUT"]
_statusT = int
_urlT = str



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Provides handling of TCC gateways."""
from __future__ import annotations

from typing import TYPE_CHECKING, Final

from .controlsystem import ControlSystem
from .exceptions import InvalidSchema
from .schema.const import (
    SZ_GATEWAY,
    SZ_GATEWAY_ID,
    SZ_GATEWAY_INFO,
    SZ_IS_WI_FI,
    SZ_MAC,
    SZ_TEMPERATURE_CONTROL_SYSTEMS,
)
from .schema.status import SCH_GATEWAY

if TYPE_CHECKING:
    import logging

    from . import Broker, Location
    from .schema import _EvoDictT, _GatewayIdT


class Gateway:
    """Instance of a location's gateway."""

    STATUS_SCHEMA: Final = SCH_GATEWAY
    TYPE: Final[str] = SZ_GATEWAY

    def __init__(self, location: Location, config: _EvoDictT) -> None:
        self.location = location  # parent

        self._broker: Broker = location._broker
        self._logger: logging.Logger = location._logger

        self._status: _EvoDictT = {}
        self._config: Final[_EvoDictT] = config[SZ_GATEWAY_INFO]

        try:
            assert self.gatewayId, "Invalid config dict"
        except AssertionError as exc:
            raise InvalidSchema(str(exc))
        self._id = self.gatewayId

        self._control_systems: list[ControlSystem] = []
        self.control_systems: dict[str, ControlSystem] = {}  # tcs by id

        tcs_config: _EvoDictT

        for tcs_config in config[SZ_TEMPERATURE_CONTROL_SYSTEMS]:
            tcs = ControlSystem(self, tcs_config)

            self._control_systems.append(tcs)
            self.control_systems[tcs.systemId] = tcs

    def __str__(self) -> str:
        return f"{self._id} ({self.TYPE})"

    @property
    def gatewayId(self) -> _GatewayIdT:
        return self._config[SZ_GATEWAY_ID]

    @property
    def mac(self) -> str:
        return self._config[SZ_MAC]

    @property
    def isWiFi(self) -> bool:
        return self._config[SZ_IS_WI_FI]

    def _update_status(self, status: _EvoDictT) -> None:
        self._status = status



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Mocked vendor server for provision via a hacked aiohttp."""
from __future__ import annotations


# normally, we want these flags to be False
_DEBUG_USE_REAL_AIOHTTP = False
_DISABLE_STRICT_ASSERTS = False  # of response content-type, schema



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohome-async - validate the config & status schemas."""
from __future__ import annotations

import json
from pathlib import Path

from evohomeasync2.schema.schedule import (
    SCH_GET_SCHEDULE_DHW,
    SCH_GET_SCHEDULE_ZONE,
    SCH_PUT_SCHEDULE_DHW,
    SCH_PUT_SCHEDULE_ZONE,
    convert_to_put_schedule,
    convert_to_get_schedule,
)


TEST_DIR = Path(__file__).resolve().parent
WORK_DIR = Path(f"{TEST_DIR}/schedules")


def _test_schedule_schema(file_name: str, schema) -> dict:
    def read_dict_from_file(file_name: str):
        with open(WORK_DIR.joinpath(file_name)) as f:
            data: dict = json.load(f)
        return data

    return schema(read_dict_from_file(file_name))


def _test_schema_schedule_dhw() -> None:
    get_sched = _test_schedule_schema("schedule_dhw_get.json", SCH_GET_SCHEDULE_DHW)
    put_sched = _test_schedule_schema("schedule_dhw_put.json", SCH_PUT_SCHEDULE_DHW)

    assert put_sched == convert_to_put_schedule(get_sched)
    assert get_sched == convert_to_get_schedule(put_sched)


def _test_schema_schedule_zone() -> None:
    get_sched = _test_schedule_schema("schedule_zone_get.json", SCH_GET_SCHEDULE_ZONE)
    put_sched = _test_schedule_schema("schedule_zone_put.json", SCH_PUT_SCHEDULE_ZONE)

    assert put_sched == convert_to_put_schedule(get_sched)
    assert get_sched == convert_to_get_schedule(put_sched)


def test_schema_schedule_dhw() -> None:
    """Test the schedule schema for dhw."""
    _test_schema_schedule_dhw()


def test_schema_schedule_zone() -> None:
    """Test the schedule schema for zones."""
    _test_schema_schedule_zone()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""evohomeasync2 provides an async client for the updated Evohome API."""
from __future__ import annotations

from http import HTTPMethod, HTTPStatus
from datetime import datetime as dt
from datetime import timedelta as td
from typing import TYPE_CHECKING, Any

import aiohttp

from .const import (
    AUTH_HEADER_ACCEPT,
    AUTH_HEADER,
    AUTH_URL,
    URL_BASE,
    AUTH_PAYLOAD,
    CREDS_REFRESH_TOKEN,
    CREDS_USER_PASSWORD,
)
from .exceptions import (
    AuthenticationFailed,
    RequestFailed,
)
from .schema import vol  # voluptuous
from .schema.account import (
    SCH_OAUTH_TOKEN,
    SZ_ACCESS_TOKEN,
    SZ_EXPIRES_IN,
    SZ_REFRESH_TOKEN,
)

if TYPE_CHECKING:
    import logging

    from .schema import _EvoDictT, _EvoListT, _EvoSchemaT


_ERR_MSG_LOOKUP_BOTH: dict[int, str] = {  # common to both OAUTH_URL & URL_BASE
    HTTPStatus.INTERNAL_SERVER_ERROR: "Can't reach server (check vendor's status page)",
    HTTPStatus.METHOD_NOT_ALLOWED: "Method not allowed (dev/test?)",
    HTTPStatus.SERVICE_UNAVAILABLE: "Can't reach server (check vendor's status page)",
    HTTPStatus.TOO_MANY_REQUESTS: "Vendor's API rate limit exceeded (wait a while)",
}

_ERR_MSG_LOOKUP_AUTH: dict[int, str] = _ERR_MSG_LOOKUP_BOTH | {  # POST OAUTH_URL
    HTTPStatus.BAD_REQUEST: "Invalid user credentials (check the username/password)",
    HTTPStatus.NOT_FOUND: "Not Found (invalid URL?)",
    HTTPStatus.UNAUTHORIZED: "Invalid access token (dev/test only)",
}

_ERR_MSG_LOOKUP_BASE: dict[int, str] = _ERR_MSG_LOOKUP_BOTH | {  # GET/PUT URL_BASE
    HTTPStatus.BAD_REQUEST: "Bad request (invalid data/json?)",
    HTTPStatus.NOT_FOUND: "Not Found (invalid entity type?)",
    HTTPStatus.UNAUTHORIZED: "Unauthorized (expired access token/unknown entity id?)",
}


class Broker:
    """A class for interacting with the Evohome API."""

    def __init__(
        self,
        username: str,
        password: str,
        /,
        *,
        refresh_token: str | None = None,
        access_token: str | None = None,
        access_token_expires: dt | None = None,
        session: aiohttp.ClientSession | None = None,
        logger: logging.Logger | None = None,
    ) -> None:
        """A class for interacting with the Evohome API."""

        self._credentials = {
            "Username": username,
            "Password": password,
        }  # NOTE: must be Uppercase keys
        self._logger = logger or logging.getLogger(__name__)

        self.refresh_token = refresh_token
        self.access_token = access_token
        self.access_token_expires = access_token_expires

        self._session = session or aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(total=30)
        )

    async def _client(
        self, method, url, data=None, json=None, headers=None
    ) -> tuple[aiohttp.ClientResponse, None | str | _EvoDictT | _EvoListT]:
        """Wrapper for aiohttp.ClientSession()."""

        if headers is None:
            headers = await self._headers()

        if method == HTTPMethod.GET:
            _session_method = self._session.get
            kwargs = {"headers": headers}

        elif method == HTTPMethod.POST:
            _session_method = self._session.post
            kwargs = {"headers": headers, "data": data}

        elif method == HTTPMethod.PUT:
            _session_method = self._session.put
            # headers["Content-Type"] = "application/json"
            kwargs = {"headers": headers, "json": json}

        async with _session_method(url, **kwargs) as response:
            if not response.content_length:
                content = None
                self._logger.info(f"{method} {url} ({response.status}) = {content}")

            elif response.content_type == "application/json":
                content = await response.json()
                self._logger.info(f"{method} {url} ({response.status}) = {content}")

            else:  # assume "text/plain" or "text/html"
                content = await response.text()
                self._logger.debug(f"{method} {url} ({response.status}) = {content}")

            return response, content  # FIXME: is messy to return response

    async def _headers(self) -> dict[str, str]:
        """Ensure the Authorization Header has a valid Access Token."""

        if not self.access_token or not self.access_token_expires:
            await self._basic_login()

        elif dt.now() > self.access_token_expires:
            await self._basic_login()

        assert isinstance(self.access_token, str)  # mypy

        return {
            "Accept": AUTH_HEADER_ACCEPT,
            "Authorization": "bearer " + self.access_token,
            "Content-Type": "application/json",
        }

    async def _basic_login(self) -> None:
        """Obtain a new access token from the vendor (as it is invalid, or expired).

        First, try using the refresh token, if one is available, otherwise authenticate
        using the user credentials.
        """

        # assert (
        #     not self.access_token
        #     or not self.access_token_expires
        #     or dt.now() > self.access_token_expires - td(seconds=30)
        # )

        self._logger.debug("No/Expired/Invalid access_token, re-authenticating.")
        self.access_token = self.access_token_expires = None

        if self.refresh_token:
            self._logger.debug("Authenticating with the refresh_token...")
            credentials = {SZ_REFRESH_TOKEN: self.refresh_token}

            try:
                await self._obtain_access_token(CREDS_REFRESH_TOKEN | credentials)  # type: ignore[arg-type]

            except AuthenticationFailed as exc:
                if exc.status != HTTPStatus.BAD_REQUEST:  # e.g. invalid tokens
                    raise

                self._logger.warning(
                    "Likely Invalid refresh_token (will try username/password)"
                )
                self.refresh_token = None

        if not self.refresh_token:
            self._logger.debug("Authenticating with username/password...")
            await self._obtain_access_token(CREDS_USER_PASSWORD | self._credentials)  # type: ignore[arg-type]

        self._logger.debug(f"refresh_token = {self.refresh_token}")
        self._logger.debug(f"access_token = {self.access_token}")
        self._logger.debug(f"access_token_expires = {self.access_token_expires}")

    async def _obtain_access_token(self, credentials: dict[str, int | str]) -> None:
        """Obtain an access token using either the refresh token or user credentials."""

        response: aiohttp.ClientResponse
        content: dict[str, int | str]

        try:
            response, content = await self._client(  # type: ignore[assignment]
                HTTPMethod.POST,
                AUTH_URL,
                data=AUTH_PAYLOAD | credentials,
                headers=AUTH_HEADER,
            )
            response.raise_for_status()

        except aiohttp.ClientResponseError as exc:
            if hint := _ERR_MSG_LOOKUP_AUTH.get(exc.status):
                raise AuthenticationFailed(hint, status=exc.status)
            raise AuthenticationFailed(str(exc), status=exc.status)

        except aiohttp.ClientError as exc:  # e.g. ClientConnectionError
            raise AuthenticationFailed(str(exc))

        try:  # the access token _should_ be valid...
            _ = SCH_OAUTH_TOKEN(content)  # can't use result, due to obsfucated values
        except vol.Invalid as exc:
            self._logger.warning(
                f"Response may be invalid (bad schema): POST {AUTH_URL}: {exc}"
            )

        try:
            self.access_token = content[SZ_ACCESS_TOKEN]  # type: ignore[assignment]
            self.access_token_expires = (
                dt.now() + td(seconds=content[SZ_EXPIRES_IN] - 15)  # type: ignore[operator]
            )
            self.refresh_token = content[SZ_REFRESH_TOKEN]  # type: ignore[assignment]

        except (KeyError, TypeError) as exc:
            raise AuthenticationFailed(f"Invalid response from server: {exc}")

    async def get(self, url: str, schema: vol.Schema | None = None) -> _EvoSchemaT:
        """"""

        response: aiohttp.ClientResponse
        content: _EvoSchemaT

        try:
            response, content = await self._client(  # type: ignore[assignment]
                HTTPMethod.GET, f"{URL_BASE}/{url}"
            )
            response.raise_for_status()

        except aiohttp.ClientResponseError as exc:
            if hint := _ERR_MSG_LOOKUP_BASE.get(exc.status):
                raise RequestFailed(hint, status=exc.status)
            raise RequestFailed(str(exc), status=exc.status)

        except aiohttp.ClientError as exc:  # e.g. ClientConnectionError
            raise RequestFailed(str(exc))

        if schema:
            try:
                return schema(content)
            except vol.Invalid as exc:
                self._logger.warning(
                    f"Response may be invalid (bad schema): GET {url}: {exc}"
                )

        return content

    async def put(
        self, url: str, json: _EvoDictT | str, schema: vol.Schema | None = None
    ) -> dict | list[dict]:  # NOTE: not _EvoSchemaT
        """"""

        response: aiohttp.ClientResponse
        content: dict[str, Any] | list[dict[str, Any]]

        if schema:
            try:
                _ = schema(json)
            except vol.Invalid as exc:
                self._logger.warning(
                    f"JSON may be invalid (bad schema): PUT {url}: {exc}"
                )

        try:
            response, content = await self._client(  # type: ignore[assignment]
                HTTPMethod.PUT, f"{URL_BASE}/{url}", json=json
            )
            response.raise_for_status()

        except aiohttp.ClientResponseError as exc:
            if hint := _ERR_MSG_LOOKUP_BASE.get(exc.status):
                raise RequestFailed(hint, status=exc.status)
            raise RequestFailed(str(exc), status=exc.status)

        except aiohttp.ClientError as exc:  # e.g. ClientConnectionError
            raise RequestFailed(str(exc))

        return content
